# Copyright 2020 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for mpo_ops.py."""

import functools
import math
from typing import Tuple

from absl.testing import absltest
from absl.testing import parameterized
import chex
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
import optax
from rlax._src import distributions
from rlax._src import mpo_ops

Array = chex.Array
Numeric = chex.Numeric

NUM_SAMPLES = 10
ACTION_DIM = 3
TIME_DIM = 8
BATCH_DIM = 100

# NOTE: These are not typical values used for MPO. In the test case, we know the
# Q function perfectly so we loosen the bound on the mean to zone in to the
# optimal policy very quickly. Similarly, we maintain a high variance to sample
# distinct actions to explore and learn from.
_INIT_TEMPERATURE = 0.2
_INIT_ALPHA_MEAN = 0.001
_INIT_ALPHA_COVARIANCE = float(1e6)

_EPSILON_BOUND = 0.01
_EPSILON_MEAN_BOUND = 10.0
_EPSILON_COVARIANCE_BOUND = 1e-12

_NUM_ITERATIONS = 5000
_TARGET_UPDATE_PERIOD = 100
_RANDOM_SEED = 42

# The offset to ensure initially the policy is not close to 0
_MEAN_OFFSET = 2.0

# The final action should optimize down to be close to 0.0
_MAX_ACTION_ERROR = 0.2
_MAX_KL_ERROR = 1e-6

_DIAGONAL_GAUSSIAN_DIST = distributions.gaussian_diagonal()
_PROJECTION_OPERATOR = functools.partial(jnp.clip, a_min=1e-10)


def _hk_mock_policy_params(s_tm1):
  """Returns mock policy params."""
  # Outputs of the network are mu and sigma. Both shaped [B, ACTION_DIM].
  pi_out = hk.nets.MLP(
      output_sizes=[2 * ACTION_DIM],
      w_init=hk.initializers.VarianceScaling(1e-3),
      activation=jnp.tanh,
      activate_final=False,
      name='online_policy')(s_tm1)
  pi_mean, pi_cov = jnp.split(pi_out, 2, axis=-1)
  pi_cov = jax.nn.softplus(pi_cov)
  pi_mean = pi_mean + _MEAN_OFFSET
  return {'mean': pi_mean, 'stddev': pi_cov}


def _init_params(key):
  init_fn, _ = hk.transform(_hk_mock_policy_params)
  key_seq = hk.PRNGSequence(key)
  s_tm1 = jax.random.normal(
      next(key_seq), (TIME_DIM, BATCH_DIM, ACTION_DIM), jnp.float32)
  online_params = init_fn(next(key_seq), s_tm1)
  return dict(
      online=online_params,
      target=online_params,
      mpo=dict(
          temperature=_INIT_TEMPERATURE,
          alpha_mean=_INIT_ALPHA_MEAN,
          alpha_covariance=_INIT_ALPHA_COVARIANCE),
  )


def _mock_outputs(online_params, target_params, key, target_name):
  """Returns mock network outputs."""
  _, policy_params_fn = hk.transform(_hk_mock_policy_params)
  key_seq = hk.PRNGSequence(key)

  state_size = ACTION_DIM

  # Input state: [TIME_DIM, BATCH_DIM, DIM_STATE]
  s_tm1 = jax.random.normal(
      next(key_seq), (TIME_DIM, BATCH_DIM, state_size), jnp.float32)
  policy_params = policy_params_fn(online_params, None, s_tm1)
  target_policy_params = policy_params_fn(target_params, None, s_tm1)

  # Shape for actions: [NUM_SAMPLES, TIME_DIM, BATCH_DIM, ACTION_DIM]
  mean, stddev = target_policy_params['mean'], target_policy_params['stddev']
  mean_repeated = jnp.repeat(
      mean.reshape((1,) + mean.shape), NUM_SAMPLES, axis=0)
  stddev_repeated = jnp.repeat(
      stddev.reshape((1,) + stddev.shape), NUM_SAMPLES, axis=0)
  target_actions = _DIAGONAL_GAUSSIAN_DIST.sample(
      next(key_seq), mean_repeated, stddev_repeated)
  # If the target is advantages then num samples is 1.
  if target_name == 'advantages':
    target_actions = target_actions[0, ...]

  # Shape for Q: [NUM_SAMPLES, TIME_DIM, BATCH_DIM]
  # Setting Q = -a_t * tf.transpose(a_t) where a_t = s_t + a.
  # The solution to optimizing this is basically for the policy to output
  # 0 actions thereby minimizing the cost. Since this is a convex
  # optimization problem, the algorithm should get to a good solution quickly.

  # First compute a_t = s_t + a with shape: [NUM_SAMPLES, TIME_DIM, BATCH_DIM,
  # ACTION_DIM] since action dim is the same as shape dim here and then compute
  # the quadratic form.
  a_t = target_actions + jnp.expand_dims(s_tm1, 0)
  sample_q_values = -jnp.sum(a_t ** 2, axis=-1)
  # Set the advantage to the same as the q value.
  # Shape for advantages: [TIME_DIM, BATCH_DIM]
  advantages = sample_q_values[0, :, :]

  return dict(
      pi_params=policy_params,
      target_pi_params=target_policy_params,
      sample_q_values=sample_q_values,
      advantages=advantages,
      target_actions=target_actions,
  )


def get_common_loss_fn_inputs(params, key, target_name):
  out = _mock_outputs(params['online'], params['target'], key, target_name)
  pi_sample_log_probs = _DIAGONAL_GAUSSIAN_DIST.logprob(
      out['target_actions'], out['pi_params']['mean'],
      out['pi_params']['stddev'])

  return out, {
      'sample_log_probs': pi_sample_log_probs,
      target_name: out[target_name],
      'temperature_constraint': mpo_ops.LagrangePenalty(
          params['mpo']['temperature'], _EPSILON_BOUND)}


def _decoupled_multivariate_normal_kl_divergence(
    mu_0: Array, sigma_0: Numeric, mu_1: Array, sigma_1: Numeric,
    per_dimension: bool = False
) -> Tuple[Array, Array]:
  """Compute the KL between diagonal Gaussians decomposed into mean and covar-e.

  Args:
    mu_0: array like of mean values for policy 0
    sigma_0: array like of std values for policy 0
    mu_1: array like of mean values for policy 1
    sigma_1: array like of std values for policy 1
    per_dimension: Whether to return a separate kl divergence for each dimension
      on the last axis.

  Returns:
    the kl divergence between the distributions decomposed into mean and
    covariance.
  """
  # Support scalar and vector `sigma`. If vector, mu.shape==sigma.shape.
  sigma_1 = jnp.ones_like(mu_1) * sigma_1
  sigma_0 = jnp.ones_like(mu_0) * sigma_0
  v1 = jnp.clip(sigma_1**2, 1e-6, 1e6)
  v0 = jnp.clip(sigma_0**2, 1e-6, 1e6)
  mu_diff = mu_1 - mu_0
  kl_mean = 0.5 * jnp.divide(mu_diff**2, v1)
  kl_cov = 0.5 * (jnp.divide(v0, v1) - jnp.ones_like(mu_1) + jnp.log(v1) -
                  jnp.log(v0))
  if not per_dimension:
    kl_mean = jnp.sum(kl_mean, axis=-1)
    kl_cov = jnp.sum(kl_cov, axis=-1)

  return kl_mean, kl_cov


def get_decoupled_kl_constraints(out, params, per_dimension):
  """Factorises KL for Gaussian."""
  kl_mean, kl_covariance = (
      _decoupled_multivariate_normal_kl_divergence(
          out['target_pi_params']['mean'], out['target_pi_params']['stddev'],
          out['pi_params']['mean'], out['pi_params']['stddev'],
          per_dimension=per_dimension))
  alpha_mean = params['mpo']['alpha_mean'] * jnp.ones_like(kl_mean)
  alpha_covariance = params['mpo']['alpha_covariance'] * jnp.ones_like(
      kl_covariance)

  return [
      (kl_mean, mpo_ops.LagrangePenalty(
          alpha=alpha_mean, epsilon=_EPSILON_MEAN_BOUND,
          per_dimension=per_dimension)),
      (kl_covariance, mpo_ops.LagrangePenalty(
          alpha=alpha_covariance, epsilon=_EPSILON_COVARIANCE_BOUND,
          per_dimension=per_dimension)),
  ]


def get_coupled_kl_constraints(out, params, per_dimension):
  kl_mean, kl_covariance = (
      _decoupled_multivariate_normal_kl_divergence(
          out['target_pi_params']['mean'], out['target_pi_params']['stddev'],
          out['pi_params']['mean'], out['pi_params']['stddev'],
          per_dimension=per_dimension))
  alpha_mean = params['mpo']['alpha_mean'] * jnp.ones_like(kl_mean)
  return [
      (kl_mean + kl_covariance, mpo_ops.LagrangePenalty(
          alpha=alpha_mean,
          epsilon=_EPSILON_MEAN_BOUND + _EPSILON_COVARIANCE_BOUND,
          per_dimension=per_dimension))
  ]


def vmpo_e_step_without_restarting_or_importance_weights(advantages, **kwargs):
  restarting_weights = jnp.ones_like(advantages)
  importance_weights = jnp.ones_like(advantages)
  return mpo_ops.vmpo_compute_weights_and_temperature_loss(
      advantages=advantages, restarting_weights=restarting_weights,
      importance_weights=importance_weights, **kwargs)


class MPOTest(parameterized.TestCase):
  """Tests for the MPO losses."""

  @parameterized.parameters(
      {'target_name': 'sample_q_values',
       'loss_fn': mpo_ops.mpo_loss,
       'get_kl_constraints': get_decoupled_kl_constraints,
       'per_dimension': False},
      {'target_name': 'advantages',
       'loss_fn': mpo_ops.vmpo_loss,
       'get_kl_constraints': get_decoupled_kl_constraints,
       'per_dimension': False},
      {'target_name': 'sample_q_values',
       'loss_fn': mpo_ops.mpo_loss,
       'get_kl_constraints': get_coupled_kl_constraints,
       'per_dimension': False},
      {'target_name': 'advantages',
       'loss_fn': mpo_ops.vmpo_loss,
       'get_kl_constraints': get_coupled_kl_constraints,
       'per_dimension': False},
      {'target_name': 'sample_q_values',
       'loss_fn': mpo_ops.mpo_loss,
       'get_kl_constraints': get_decoupled_kl_constraints,
       'per_dimension': True},
      {'target_name': 'advantages',
       'loss_fn': mpo_ops.vmpo_loss,
       'get_kl_constraints': get_decoupled_kl_constraints,
       'per_dimension': True},
      {'target_name': 'sample_q_values',
       'loss_fn': mpo_ops.mpo_loss,
       'get_kl_constraints': get_coupled_kl_constraints,
       'per_dimension': True},
      {'target_name': 'advantages',
       'loss_fn': mpo_ops.vmpo_loss,
       'get_kl_constraints': get_coupled_kl_constraints,
       'per_dimension': True},
  )
  def test_optimization(
      self, target_name, loss_fn, get_kl_constraints, per_dimension):
    """Tests that the policy optimization works correctly."""

    def _loss(params, key):
      out, loss_fn_inputs = get_common_loss_fn_inputs(params, key, target_name)
      kl_constraints = get_kl_constraints(out, params, per_dimension)
      loss_fn_inputs.update({'kl_constraints': kl_constraints})
      loss, mpo_stats = loss_fn(**loss_fn_inputs)
      loss = jnp.mean(loss)
      temperature_bound = jnp.mean(mpo_stats.normalized_weights * jnp.log(
          mpo_stats.num_samples * mpo_stats.normalized_weights + 1e-8))
      return loss, {'outputs': out, 'temperature_bound': temperature_bound}

    key = jax.random.PRNGKey(_RANDOM_SEED)
    grad_fn = jax.jit(jax.grad(_loss, has_aux=True))
    optimizer = optax.adam(1e-3)
    key, new_key = jax.random.split(key)
    params = _init_params(new_key)
    opt_state = optimizer.init((params['online'], params['mpo']))

    @jax.jit
    def _update(params_, opt_state_, key_):
      next_key, key_ = jax.random.split(key_)
      grad, stats = grad_fn(params_, key_)
      updates, opt_state_ = optimizer.update(
          (grad['online'], grad['mpo']), opt_state_)
      online_params, mpo_params = optax.apply_updates(
          (params_['online'], params_['mpo']), updates)
      params_['online'] = online_params
      params_['mpo'] = mpo_params
      return params_, opt_state_, stats, next_key

    for iter_idx in range(_NUM_ITERATIONS):
      params, opt_state, extra, key = _update(params, opt_state, key)
      if iter_idx % _TARGET_UPDATE_PERIOD == 0:
        params['target'] = params['online']

    # Test the bounds are within tolerance.
    key, new_key = jax.random.split(key)
    _, extra = _loss(params, new_key)
    action_mean = jnp.mean(extra['outputs']['pi_params']['mean'])
    # Check action mean is close to 0.
    self.assertBetween(action_mean, -_MAX_ACTION_ERROR, _MAX_ACTION_ERROR)

    # Check the temperature are within the bounds.
    self.assertLess(extra['temperature_bound'], _EPSILON_BOUND)

  @parameterized.parameters(
      {'e_step_fn': mpo_ops.mpo_compute_weights_and_temperature_loss,
       'additional_inputs': {},
       # dL/dq == 1 and dL/dt == epsilon (for one sample)
       'expected_deriv_of_target': [[[1]]],
       'sample_dimension': True},
      {'e_step_fn': vmpo_e_step_without_restarting_or_importance_weights,
       'additional_inputs': {'top_k_fraction': 1.0},
       'expected_deriv_of_target': [[1]],
       'sample_dimension': False},
  )
  def test_e_step_gradient_computation(
      self, e_step_fn, additional_inputs, expected_deriv_of_target,
      sample_dimension):
    """Tests the gradients from the E-step against the analytic ones."""
    # Target has shape [NUM_SAMPLES, T, B] => [1, 1, 1]
    target = jnp.array([[3]], jnp.float32)
    if sample_dimension:
      target = jnp.expand_dims(target, axis=0)
    temperature = jnp.array(0.1, jnp.float32)
    def fn(target_, temperature_):
      temperature_constraint = mpo_ops.LagrangePenalty(
          temperature_, _EPSILON_BOUND)
      temperature_loss, _, _ = e_step_fn(
          target_, temperature_constraint=temperature_constraint,
          projection_operator=_PROJECTION_OPERATOR,
          **additional_inputs)
      return jnp.mean(temperature_loss)
    grad = jax.grad(fn, argnums=(0, 1))(target, temperature)

    np.testing.assert_almost_equal(np.array(grad[0]), np.array(
        expected_deriv_of_target, np.float32), decimal=4)
    self.assertAlmostEqual(grad[1], _EPSILON_BOUND, places=4)

  @parameterized.parameters(
      {'e_step_fn': mpo_ops.mpo_compute_weights_and_temperature_loss,
       'additional_inputs': {},
       'sample_dimension': True},
      {'e_step_fn': vmpo_e_step_without_restarting_or_importance_weights,
       'additional_inputs': {'top_k_fraction': 1.0},
       'sample_dimension': False},
  )
  def test_e_step_stop_gradient(
      self, e_step_fn, additional_inputs, sample_dimension):
    """Tests no gradients flow through `weights` in the E-Step."""
    # Target has shape [NUM_SAMPLES, T, B] => [1, 1, 1]
    target = jnp.array([[3]], jnp.float32)
    if sample_dimension:
      target = jnp.expand_dims(target, axis=0)
    temperature = 0.1
    # pylint: disable=g-long-lambda
    def mean_weights_fn(target_, temperature_):
      temperature_constraint = mpo_ops.LagrangePenalty(
          temperature_, _EPSILON_BOUND)
      _, weights, _ = e_step_fn(
          target_, temperature_constraint=temperature_constraint,
          projection_operator=_PROJECTION_OPERATOR,
          **additional_inputs)
      return jnp.mean(weights)
    grad = jax.grad(mean_weights_fn, argnums=(0, 1))(target, temperature)
    np.testing.assert_almost_equal(
        np.array(grad[0]), np.zeros_like(grad[0]), decimal=4)
    self.assertAlmostEqual(grad[1], 0., places=4)

  def test_kl_constraint_loss_gradients(self):
    """Tests the gradients in the `_kl_constraint_loss` method."""
    kl = jnp.array(1., jnp.float32)
    alpha = jnp.array(1., jnp.float32)
    _, _, alpha = mpo_ops.kl_constraint_loss(kl, mpo_ops.LagrangePenalty(
        alpha=alpha, epsilon=_EPSILON_MEAN_BOUND, per_dimension=False),
                                             _PROJECTION_OPERATOR)

    def alpha_loss_fn(alpha_):
      penalty = mpo_ops.LagrangePenalty(
          alpha=alpha_, epsilon=_EPSILON_MEAN_BOUND, per_dimension=False)
      _, alpha_loss, _ = mpo_ops.kl_constraint_loss(
          kl, penalty, _PROJECTION_OPERATOR)
      return alpha_loss
    alpha_gradients = jax.grad(alpha_loss_fn)(alpha)
    actual_alpha_gradients = _EPSILON_MEAN_BOUND - kl

    def kl_loss_fn(kl_):
      penalty = mpo_ops.LagrangePenalty(
          alpha=alpha, epsilon=_EPSILON_MEAN_BOUND, per_dimension=False)
      kl_loss, _, _ = mpo_ops.kl_constraint_loss(
          kl_, penalty, _PROJECTION_OPERATOR)
      return kl_loss
    kl_gradients = jax.grad(kl_loss_fn)(kl)
    actual_kl_gradients = alpha

    self.assertAlmostEqual(kl_gradients, actual_kl_gradients)
    self.assertAlmostEqual(alpha_gradients, actual_alpha_gradients)

  def test_kl_constraint_loss_stop_gradients(self):
    """Tests the stop gradients in the `kl_constraint_loss` function.

      The `alpha_loss` term should not affect the KL and the `kl` term should
      not affect `alpha`.
    """
    kl = jnp.array(1., jnp.float32)
    alpha = jnp.array(1., jnp.float32)
    _, _, alpha = mpo_ops.kl_constraint_loss(kl, mpo_ops.LagrangePenalty(
        alpha=alpha, epsilon=_EPSILON_MEAN_BOUND, per_dimension=False),
                                             _PROJECTION_OPERATOR)

    def kl_loss_fn(alpha_):
      penalty = mpo_ops.LagrangePenalty(
          alpha=alpha_, epsilon=_EPSILON_MEAN_BOUND, per_dimension=False)
      kl_loss, _, _ = mpo_ops.kl_constraint_loss(
          kl, penalty, _PROJECTION_OPERATOR)
      return kl_loss

    kl_gradients = jax.grad(kl_loss_fn)(alpha)

    def alpha_loss_fn(kl_):
      penalty = mpo_ops.LagrangePenalty(
          alpha=alpha, epsilon=_EPSILON_MEAN_BOUND, per_dimension=False)
      _, alpha_loss, _ = mpo_ops.kl_constraint_loss(
          kl_, penalty, _PROJECTION_OPERATOR)
      return alpha_loss
    alpha_gradients = jax.grad(alpha_loss_fn)(kl)

    # Test that there are no gradients of KL w.r.t alpha
    self.assertEqual(kl_gradients, 0.)

    # Test that there are no gradients of alpha w.r.t kl
    self.assertEqual(alpha_gradients, 0.)

  @parameterized.parameters(
      # With restarting weights of 1 (and temperature of 1) the weights should
      # be e^-1, 1, max advantage is 2 and num samples is 2 so temperature loss
      # is log(1 + e^-1) + 2 - log(2) + temperature epsilon
      {'advantages': np.array([[1.0, 2.0]]),
       'restarting_weights': np.array([[1.0, 1.0]]),
       'expected_temperature_loss': (math.log(1.0 + math.exp(-1.0)) + 2.0 -
                                     math.log(2.0) + _EPSILON_BOUND)},
      # With the second restarting weight set to 0 the weights become 1, 0
      # max advantage is 1 and num samples is 1 so temperature loss is
      # log(1) + 1 - log(1) + temperature epsilon
      {'advantages': np.array([[1.0, 2.0]]),
       'restarting_weights': np.array([[1.0, 0.0]]),
       'expected_temperature_loss': 1.0 + _EPSILON_BOUND},
  )
  def test_restarting_weights(
      self, advantages, restarting_weights, expected_temperature_loss):
    """Test that calculation is correct if restarting weight is set to 0."""
    temperature_loss, _, _ = mpo_ops.vmpo_compute_weights_and_temperature_loss(
        advantages, restarting_weights, np.ones_like(restarting_weights),
        mpo_ops.LagrangePenalty(1.0, _EPSILON_BOUND),
        functools.partial(np.clip, a_min=1e-8, a_max=None), 1.0)
    self.assertAlmostEqual(
        temperature_loss, expected_temperature_loss, places=4)

  @parameterized.parameters(
      # When the top k fraction is 1.0 all of the weights should be 1
      {'top_k_fraction': 1.0,
       'scaled_advantages': np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
       'expected_top_k_weights': np.array([[1.0, 1.0, 1.0], [1.0, 1.0, 1.0]])},
      # When the top k fraction is 0.5 it will take the bottom row as these are
      # the highest.
      {'top_k_fraction': 0.5,
       'scaled_advantages': np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]),
       'expected_top_k_weights': np.array([[0.0, 0.0, 0.0], [1.0, 1.0, 1.0]])}
  )
  def test_top_k_fraction(
      self, top_k_fraction, scaled_advantages, expected_top_k_weights):
    """Test that only the top k fraction are used."""
    top_k_weights = mpo_ops.get_top_k_weights(
        top_k_fraction, jnp.ones_like(scaled_advantages), scaled_advantages)
    np.testing.assert_allclose(top_k_weights, expected_top_k_weights)

  def test_top_k_fraction_too_low(self):
    """Test if the top k fraction returns 0 advantages we raise an error."""
    with self.assertRaises(ValueError):
      mpo_ops.get_top_k_weights(0.01, jnp.ones((3, 2)), jnp.ones((3, 2)))

  @parameterized.parameters(
      # With importance weights of 1 (and temperature of 1) the weights should
      # be e^-1, 1, max advantage is 2 and num samples is 2 so temperature loss
      # is log(1 + e^-1) + 2 - log(2) + temperature epsilon
      {'advantages': np.array([[1.0, 2.0]]),
       'importance_weights': np.array([[1.0, 1.0]]),
       'expected_temperature_loss': (math.log(1.0 + math.exp(-1.0)) + 2.0 -
                                     math.log(2.0) + _EPSILON_BOUND)},
      # If the second importance weight is 0.5 temperature loss becomes
      # log(0.5 + e^-1) + 2 - log(2) + temperature epsilon
      {'advantages': np.array([[1.0, 2.0]]),
       'importance_weights': np.array([[1.0, 0.5]]),
       'expected_temperature_loss': (math.log(0.5 + math.exp(-1.0)) + 2.0 -
                                     math.log(2.0) + _EPSILON_BOUND)},
  )
  def test_importance_weights(
      self, advantages, importance_weights, expected_temperature_loss):
    """Test that importance weights have the correct effect."""
    temperature_loss, _, _ = mpo_ops.vmpo_compute_weights_and_temperature_loss(
        advantages, np.ones_like(importance_weights), importance_weights,
        mpo_ops.LagrangePenalty(1.0, _EPSILON_BOUND),
        functools.partial(np.clip, a_min=1e-8, a_max=None), 1.0)
    self.assertAlmostEqual(
        temperature_loss, expected_temperature_loss, places=4)

  @parameterized.parameters({'per_dimension': True}, {'per_dimension': False})
  def test_mpo_input_axis_order_equivalence(self, per_dimension):
    """Test loss functions are equivalent regardless of axis order."""
    key = jax.random.PRNGKey(_RANDOM_SEED)
    key, new_key = jax.random.split(key)
    params = _init_params(new_key)
    out, mpo_inputs = get_common_loss_fn_inputs(params, key, 'sample_q_values')
    kl_constraints = get_coupled_kl_constraints(out, params,
                                                per_dimension=per_dimension)
    mpo_inputs.update({'kl_constraints': kl_constraints})

    # Original loss fn inputs are [S T B],
    stb_loss, stb_outputs = mpo_ops.mpo_loss(**mpo_inputs)
    mean_stb_loss = jnp.mean(stb_loss)

    # Swap axes and try [S B T]
    mpo_inputs.update({
        'sample_log_probs': jnp.swapaxes(mpo_inputs['sample_log_probs'], 1, 2),
        'sample_q_values': jnp.swapaxes(mpo_inputs['sample_q_values'], 1, 2),
        'kl_constraints': [(jnp.swapaxes(kl, 0, 1), mpo_ops.LagrangePenalty(
            alpha=jnp.swapaxes(pen.alpha, 0, 1), epsilon=pen.epsilon,
            per_dimension=pen.per_dimension)) for (kl, pen) in kl_constraints],
    })
    sbt_loss, sbt_outputs = mpo_ops.mpo_loss(**mpo_inputs)
    mean_sbt_loss = jnp.mean(sbt_loss)

    # Try [T B S] denoting sample_axis at 2 instead of 0.
    mpo_inputs.update({
        'sample_log_probs': jnp.swapaxes(mpo_inputs['sample_log_probs'], 0, 2),
        'sample_q_values': jnp.swapaxes(mpo_inputs['sample_q_values'], 0, 2),
        'kl_constraints': kl_constraints,  # T B
        'sample_axis': 2
    })
    tbs_loss, tbs_outputs = mpo_ops.mpo_loss(**mpo_inputs)
    mean_tbs_loss = jnp.mean(tbs_loss)

    self.assertAlmostEqual(mean_stb_loss, mean_sbt_loss, places=4)
    self.assertAlmostEqual(mean_tbs_loss, mean_sbt_loss, places=4)
    self.assertEqual(tbs_outputs.num_samples, sbt_outputs.num_samples)
    self.assertEqual(tbs_outputs.num_samples, stb_outputs.num_samples)

  @parameterized.parameters({'per_dimension': True}, {'per_dimension': False})
  def test_vmpo_input_axis_order_equivalence(self, per_dimension):
    """Test loss functions are equivalent regardless of axis order."""
    key = jax.random.PRNGKey(_RANDOM_SEED)
    key, new_key = jax.random.split(key)
    params = _init_params(new_key)
    out, vmpo_inputs = get_common_loss_fn_inputs(params, key, 'advantages')
    kl_constraints = get_coupled_kl_constraints(out, params,
                                                per_dimension=per_dimension)
    vmpo_inputs.update({'kl_constraints': kl_constraints})

    # Original loss fn inputs are [T B],
    tb_loss, tb_outputs = mpo_ops.vmpo_loss(**vmpo_inputs)
    mean_tb_loss = jnp.mean(tb_loss)

    # Swap axes and try [B T]
    vmpo_inputs.update({
        'sample_log_probs': jnp.swapaxes(vmpo_inputs['sample_log_probs'], 0, 1),
        'advantages': jnp.swapaxes(vmpo_inputs['advantages'], 0, 1),
        'kl_constraints': [(jnp.swapaxes(kl, 0, 1), mpo_ops.LagrangePenalty(
            alpha=jnp.swapaxes(pen.alpha, 0, 1), epsilon=pen.epsilon,
            per_dimension=pen.per_dimension)) for (kl, pen) in kl_constraints],
    })
    bt_loss, bt_outputs = mpo_ops.vmpo_loss(**vmpo_inputs)
    mean_bt_loss = jnp.mean(bt_loss)

    self.assertAlmostEqual(mean_tb_loss, mean_bt_loss, places=4)
    self.assertEqual(tb_outputs.num_samples, bt_outputs.num_samples)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A simple online Q-learning agent trained to play BSuite's Catch env."""

import collections
from absl import app
from absl import flags
from bsuite.environments import catch
import haiku as hk
from haiku import nets
import jax
import jax.numpy as jnp
import optax
import rlax
from rlax.examples import experiment

ActorOutput = collections.namedtuple("ActorOutput", "actions q_values")
Transition = collections.namedtuple("Transition",
                                    "obs_tm1 a_tm1 r_t discount_t obs_t")

FLAGS = flags.FLAGS
flags.DEFINE_integer("seed", 42, "Random seed.")
flags.DEFINE_integer("train_episodes", 500, "Number of train episodes.")
flags.DEFINE_integer("num_hidden_units", 50, "Number of network hidden units.")
flags.DEFINE_float("epsilon", 0.01, "Epsilon-greedy exploration probability.")
flags.DEFINE_float("discount_factor", 0.99, "Q-learning discount factor.")
flags.DEFINE_float("learning_rate", 0.005, "Optimizer learning rate.")
flags.DEFINE_integer("eval_episodes", 100, "Number of evaluation episodes.")
flags.DEFINE_integer("evaluate_every", 50,
                     "Number of episodes between evaluations.")


def build_network(num_hidden_units: int, num_actions: int) -> hk.Transformed:
  """Factory for a simple MLP network for approximating Q-values."""

  def q(obs):
    flatten = lambda x: jnp.reshape(x, (-1,))
    network = hk.Sequential(
        [flatten, nets.MLP([num_hidden_units, num_actions])])
    return network(obs)

  return hk.without_apply_rng(hk.transform(q))


class TransitionAccumulator:
  """Simple Python accumulator for transitions."""

  def __init__(self):
    self._prev = None
    self._action = None
    self._latest = None

  def push(self, env_output, action):
    self._prev = self._latest
    self._action = action
    self._latest = env_output

  def sample(self, batch_size):
    assert batch_size == 1
    return Transition(self._prev.observation, self._action, self._latest.reward,
                      self._latest.discount, self._latest.observation)

  def is_ready(self, batch_size):
    assert batch_size == 1
    return self._prev is not None


class OnlineQ:
  """An online Q-learning deep RL agent."""

  def __init__(self, observation_spec, action_spec, num_hidden_units, epsilon,
               learning_rate):
    self._observation_spec = observation_spec
    self._action_spec = action_spec
    self._epsilon = epsilon
    # Neural net and optimiser.
    self._network = build_network(num_hidden_units, action_spec.num_values)
    self._optimizer = optax.adam(learning_rate)
    # Jitting for speed.
    self.actor_step = jax.jit(self.actor_step)
    self.learner_step = jax.jit(self.learner_step)

  def initial_params(self, key):
    sample_input = self._observation_spec.generate_value()
    return self._network.init(key, sample_input)

  def initial_actor_state(self):
    return ()

  def initial_learner_state(self, params):
    return self._optimizer.init(params)

  def actor_step(self, params, env_output, actor_state, key, evaluation):
    q = self._network.apply(params, env_output.observation)
    train_a = rlax.epsilon_greedy(self._epsilon).sample(key, q)
    eval_a = rlax.greedy().sample(key, q)
    a = jax.lax.select(evaluation, eval_a, train_a)
    return ActorOutput(actions=a, q_values=q), actor_state

  def learner_step(self, params, data, learner_state, unused_key):
    dloss_dtheta = jax.grad(self._loss)(params, *data)
    updates, learner_state = self._optimizer.update(dloss_dtheta, learner_state)
    params = optax.apply_updates(params, updates)
    return params, learner_state

  def _loss(self, params, obs_tm1, a_tm1, r_t, discount_t, obs_t):
    q_tm1 = self._network.apply(params, obs_tm1)
    q_t = self._network.apply(params, obs_t)
    td_error = rlax.q_learning(q_tm1, a_tm1, r_t, discount_t, q_t)
    return rlax.l2_loss(td_error)


def main(unused_arg):
  env = catch.Catch(seed=FLAGS.seed)
  agent = OnlineQ(
      observation_spec=env.observation_spec(),
      action_spec=env.action_spec(),
      num_hidden_units=FLAGS.num_hidden_units,
      epsilon=FLAGS.epsilon,
      learning_rate=FLAGS.learning_rate,
  )

  accumulator = TransitionAccumulator()
  experiment.run_loop(
      agent=agent,
      environment=env,
      accumulator=accumulator,
      seed=FLAGS.seed,
      batch_size=1,
      train_episodes=FLAGS.train_episodes,
      evaluate_every=FLAGS.evaluate_every,
      eval_episodes=FLAGS.eval_episodes,
  )


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for losses.py."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import losses


class L2LossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.xs = jnp.array([-2, -1, -0.5, 0, 0.5, 1, 2])
    self.ys = jnp.array([2., 0.5, 0.125, 0, 0.125, 0.5, 2.])
    self.dys = jnp.array([-2, -1, -0.5, 0, 0.5, 1, 2])

  @chex.all_variants()
  def test_l2_loss_scalar(self):
    l2_loss = self.variant(losses.l2_loss)
    x = jnp.array(0.5)
    # Test output.
    np.testing.assert_allclose(l2_loss(x), 0.125)

  @chex.all_variants()
  def test_l2_loss_vector(self):
    l2_loss = self.variant(losses.l2_loss)
    # Test output.
    np.testing.assert_allclose(l2_loss(self.xs), self.ys)

  @chex.all_variants()
  def test_l2_regularizer(self):
    l2_loss = self.variant(losses.l2_loss)
    # Test output.
    np.testing.assert_allclose(
        l2_loss(self.xs), l2_loss(self.xs, jnp.zeros_like(self.xs)))

  @chex.all_variants()
  def test_gradients(self):
    l2_loss = self.variant(losses.l2_loss)
    # Compute gradient in batch
    batch_grad_func = jax.vmap(jax.grad(l2_loss), (0))
    actual = batch_grad_func(self.xs)
    np.testing.assert_allclose(actual, self.dys)


class ExpectileLossTest(parameterized.TestCase):

  @chex.all_variants()
  @parameterized.named_parameters(
      ('expectile_0.5', 0.5, [-2., -1., -0.5, 0., 0.5, 1., 2.
                             ], [2., 0.5, 0.125, 0., 0.125, 0.5, 2.]),
      ('expectile_0.0', 0.0, [-2., -1., -0.5, 0., 0.5, 1., 2.
                             ], [0., 0., 0., 0., 0.25, 1.0, 4.0]),
      ('expectile_1.0', 1.0, [-2., -1., -0.5, 0., 0.5, 1., 2.
                             ], [4.0, 1.0, 0.25, 0., 0.0, 0.0, 0.0]),
      ('expectile_0.75', 0.75, [-2., -1., -0.5, 0., 0.5, 1., 2.
                               ], [3.0, 0.75, 0.1875, 0., 0.0625, 0.25, 1.0]),
  )
  def test_expectile_loss_vector(self, expectile, predictions, expected_loss):
    expectile_loss = self.variant(losses.expectile_loss)
    predictions = jnp.array(predictions)
    expected_loss = jnp.array(expected_loss)
    targets = jnp.zeros_like(predictions)
    # Test output.
    np.testing.assert_allclose(
        expectile_loss(
            predictions=predictions,
            targets=targets,
            expectile=expectile,
        ),
        expected_loss,
        atol=1e-4)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('expectile_0.5', 0.5, [-2., -1., -0.5, 0., 0.5, 1., 2.
                             ], [-2., -1., -0.5, 0., 0.5, 1.0, 2.0]),
      ('expectile_0.0', 0.0, [-2., -1., -0.5, 0., 0.5, 1., 2.
                             ], [0., 0., 0., 0., 1.0, 2.0, 4.0]),
      ('expectile_1.0', 1.0, [-2., -1., -0.5, 0., 0.5, 1., 2.
                             ], [-4.0, -2.0, -1.0, 0.0, 0.0, 0.0, 0.0]),
      ('expectile_0.75', 0.75, [-2., -1., -0.5, 0., 0.5, 1., 2.
                               ], [-3.0, -1.5, -0.75, 0., 0.25, 0.50, 1.0]),
  )
  def test_gradients(self, expectile, predictions, expected_grads):
    predictions = jnp.array(predictions)
    expected_grads = jnp.array(expected_grads)
    targets = jnp.zeros_like(predictions)
    expectile_loss_fn = self.variant(
        functools.partial(losses.expectile_loss, expectile=expectile))
    batch_grad_fn = jax.vmap(jax.grad(expectile_loss_fn), (0, 0))
    predicted_grads = batch_grad_fn(predictions, targets)
    np.testing.assert_allclose(predicted_grads, expected_grads, atol=1e-4)


class LogLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.preds = jnp.array([1., 1., 0., 0., 0.5, 0.5])
    self.targets = jnp.array([1., 0., 0., 1., 1., 0])
    self.expected = jnp.array([0., np.inf, 0., np.inf, 0.6931472, 0.6931472])

  @chex.all_variants()
  def test_log_loss_scalar(self):
    log_loss = self.variant(losses.log_loss)
    preds = self.preds[2]
    targets = self.targets[2]
    # Test output.
    np.testing.assert_allclose(
        log_loss(preds, targets), self.expected[2], atol=1e-4)

  @chex.all_variants()
  def test_log_loss_vector(self):
    log_loss = self.variant(losses.log_loss)
    # Test output.
    np.testing.assert_allclose(
        log_loss(self.preds, self.targets), self.expected, atol=1e-4)


class PixelControlLossTest(parameterized.TestCase):
  """Test the `pixel_control_loss` op."""

  def setUp(self):
    """Defines example data and expected result for the op."""
    super().setUp()

    # Observation shape is (2,2,3) (i.e., height 2, width 2, and 3 channels).
    # We will use no cropping, and a cell size of 1. We have num_actions = 3,
    # meaning our Q values should be (2,2,3). We will set the Q value equal to
    # the observation.
    self.seq_length = 3
    self.discount = 0.9
    self.cell_size = 1

    # Observations.
    obs1 = np.array([[[1, 2, 3], [3, 4, 5]], [[5, 6, 7], [7, 8, 9]]])
    obs2 = np.array([[[7, 8, 9], [1, 2, 3]], [[3, 4, 5], [5, 6, 7]]])
    obs3 = np.array([[[5, 6, 7], [7, 8, 9]], [[1, 2, 3], [3, 4, 5]]])
    obs4 = np.array([[[3, 4, 5], [5, 6, 7]], [[7, 8, 9], [1, 2, 3]]])

    # Actions.
    action1 = 0
    action2 = 1
    action3 = 2

    # Compute loss for constant discount.
    qa_tm1 = obs3[:, :, action3]
    reward3 = np.mean(np.abs(obs4 - obs3), axis=2)
    qmax_t = np.amax(obs4, axis=2)
    target = reward3 + self.discount * qmax_t
    error3 = target - qa_tm1

    qa_tm1 = obs2[:, :, action2]
    reward2 = np.mean(np.abs(obs3 - obs2), axis=2)
    target = reward2 + self.discount * target
    error2 = target - qa_tm1

    qa_tm1 = obs1[:, :, action1]
    reward1 = np.mean(np.abs(obs2 - obs1), axis=2)
    target = reward1 + self.discount * target
    error1 = target - qa_tm1

    # Compute loss for episode termination with discount 0.
    qa_tm1 = obs1[:, :, action1]
    reward1 = np.mean(np.abs(obs2 - obs1), axis=2)
    target = reward1 + 0. * target
    error1_term = target - qa_tm1

    self.error = np.sum(
        np.square(error1) + np.square(error2) + np.square(error3)) * 0.5
    self.error_term = np.sum(
        np.square(error1_term) + np.square(error2) + np.square(error3)) * 0.5

    self.observations = np.stack([obs1, obs2, obs3, obs4], axis=0).astype(
        np.float32)
    self.action_values = self.observations
    self.actions = np.array([action1, action2, action3])

  @chex.all_variants()
  def testPixelControlLossScalarDiscount(self):
    """Compute loss for given observations, actions, values, scalar discount."""
    loss_fn = self.variant(functools.partial(
        losses.pixel_control_loss, cell_size=self.cell_size))
    loss = loss_fn(
        self.observations,
        self.actions,
        self.action_values,
        self.discount)
    loss = jnp.sum(loss)
    np.testing.assert_allclose(loss, self.error, rtol=1e-3)

  @chex.all_variants()
  def testPixelControlLossTensorDiscount(self):
    """Compute loss for given observations, actions, values, tensor discount."""
    zero_discount = np.zeros((1,))
    non_zero_discount = self.discount * np.ones(self.seq_length - 1)
    discount = np.concatenate([zero_discount, non_zero_discount], axis=0)
    loss_fn = self.variant(functools.partial(
        losses.pixel_control_loss, cell_size=self.cell_size))
    loss = loss_fn(
        self.observations,
        self.actions,
        self.action_values,
        discount)
    loss = jnp.sum(loss)
    np.testing.assert_allclose(loss, self.error_term, rtol=1e-3)

  @chex.all_variants()
  def testPixelControlLossShapes(self):
    with self.assertRaisesRegex(
        ValueError, 'Pixel Control values are not compatible'):
      loss_fn = self.variant(functools.partial(
          losses.pixel_control_loss, cell_size=self.cell_size))
      loss_fn(
          self.observations, self.actions,
          self.action_values[:, :-1], self.discount)

  @chex.all_variants()
  def testTensorDiscountShape(self):
    with self.assertRaisesRegex(
        ValueError, 'discount_factor must be a scalar or a tensor of rank 1'):
      discount = np.tile(
          np.reshape(self.discount, (1, 1)), (self.seq_length, 1))
      loss_fn = self.variant(functools.partial(
          losses.pixel_control_loss, cell_size=self.cell_size))
      loss_fn(
          self.observations, self.actions, self.action_values, discount)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Transformed value functions.

Canonical value functions map states onto the expected discounted sum of rewards
that may be collected by an agent from any starting state. Value functions may
also be defined as the fixed points of certain linear recursive relations known
as Bellman equations. It is sometimes useful to consider transformed values that
are the solution to non-linear generalization of traditional Bellman equations.
In this subpackage we provide a general utility for wrapping bootstrapped return
calculations to construct regression targets for these transformed values.
We also use this to implement different learning algorithms from the literature.
"""

import collections
import functools

import chex
import jax.numpy as jnp
from rlax._src import base
from rlax._src import multistep
from rlax._src import transforms

Array = chex.Array
TxPair = collections.namedtuple('TxPair', ['apply', 'apply_inv'])


# Example transform pairs; these typically consist of a monotonically increasing
# squashing fn `apply` and its inverse `apply_inv`. Other choices are possible.

IDENTITY_PAIR = TxPair(
    transforms.identity, transforms.identity)
SIGNED_LOGP1_PAIR = TxPair(
    transforms.signed_logp1, transforms.signed_expm1)
SIGNED_HYPERBOLIC_PAIR = TxPair(
    transforms.signed_hyperbolic, transforms.signed_parabolic)
HYPERBOLIC_SIN_PAIR = TxPair(
    transforms.hyperbolic_arcsin, transforms.hyperbolic_sin)
DISCOUNT_TRANSFORM_PAIR = TxPair(
    lambda x: -jnp.log(1 - x),
    lambda x: 1 - jnp.exp(-x))


def twohot_pair(
    min_value: float,
    max_value: float,
    num_bins: int) -> TxPair:
  """Construct a TxPair matching a 2-hot reparametrisation and its inverse."""
  apply_fn = functools.partial(
      transforms.transform_to_2hot,
      min_value=min_value,
      max_value=max_value,
      num_bins=num_bins)
  apply_inv_fn = functools.partial(
      transforms.transform_from_2hot,
      min_value=min_value,
      max_value=max_value,
      num_bins=num_bins)
  return TxPair(apply_fn, apply_inv_fn)


def compose_tx(*tx_list):
  """Utility to compose a sequence of TxPairs.

  The transformations are applied in order during the `apply` method:
  e.g. [f, g] --> y = g(f(x))
  and in reverse order during the `apply_inv` method:
  e.g. [f, g] --> x = f^-1(g^-1(x))

  Args:
    *tx_list: a sequence of TxPairs as positional arguments.

  Returns:
    a new TxPair.
  """

  def apply_fn(x: chex.Array):
    for tx in tx_list:
      x = tx.apply(x)
    return x

  def apply_inv_fn(x: chex.Array):
    for tx in tx_list[::-1]:
      x = tx.apply_inv(x)
    return x

  return TxPair(apply_fn, apply_inv_fn)


def muzero_pair(
    min_value: float,
    max_value: float,
    num_bins: int,
    tx: TxPair) -> TxPair:
  """Create the transformation pair introduced in MuZero.

  This more complex pair of transformations, combines a monotonic squashing
  function with a reparametrisation of the value over a fixed support.

  See "Mastering Atari, Go, Chess and Shogi by Planning with a Learned Model"
  by Schrittwieser et al. (https://arxiv.org/abs/1911.08265).

  Args:
    min_value: minimum value of the discrete support.
    max_value: maximum value of the discrete support.
    num_bins: number of discrete bins used by the fixed support.
    tx: non-linear transformation to be applied before reparametrisation.

  Returns:
    a transformation pair.
  """
  return compose_tx(tx, twohot_pair(min_value, max_value, num_bins))


def unbiased_transform_pair(
    min_value: float,
    max_value: float,
    num_bins: int,
    tx: TxPair) -> TxPair:
  """Create an unbiased value transformation pair.

  A pair of transformations, similar to 'muzero_pair' but where the
  non-linearity is applied to the bin locations rather than the scalar value.

  Args:
    min_value: minimum value of the discrete support in the transformed space.
    max_value: maximum value of the discrete support in the transformed space.
    num_bins: number of discrete bins used by the fixed support.
    tx: non-linear transformation to be applied to the bin locations.

  Returns:
    a transformation pair.
  """
  bins = tx.apply_inv(jnp.linspace(min_value, max_value, num_bins))

  apply_fn = functools.partial(
      transforms.transform_to_2hot_nonlinear, bins=bins)
  apply_inv_fn = functools.partial(
      transforms.transform_from_2hot_nonlinear, bins=bins)

  nonlinear_twohot_pair = TxPair(apply_fn, apply_inv_fn)

  return compose_tx(IDENTITY_PAIR, nonlinear_twohot_pair)


def transform_values(build_targets, *value_argnums):
  """Decorator to convert targets to use transformed value function."""
  @functools.wraps(build_targets)
  def wrapped_build_targets(tx_pair, *args, **kwargs):
    tx_args = list(args)
    for index in value_argnums:
      tx_args[index] = tx_pair.apply_inv(tx_args[index])

    targets = build_targets(*tx_args, **kwargs)
    return tx_pair.apply(targets)

  return wrapped_build_targets


transformed_lambda_returns = transform_values(multistep.lambda_returns, 2)
transformed_general_off_policy_returns_from_action_values = transform_values(
    multistep.general_off_policy_returns_from_action_values, 0)
transformed_n_step_returns = transform_values(
    multistep.n_step_bootstrapped_returns, 2)


def transformed_q_lambda(
    q_tm1: Array,
    a_tm1: Array,
    r_t: Array,
    discount_t: Array,
    q_t: Array,
    lambda_: Array,
    stop_target_gradients: bool = True,
    tx_pair: TxPair = IDENTITY_PAIR,
) -> Array:
  """Calculates Peng's or Watkins' Q(lambda) temporal difference error.

  See "General non-linear Bellman equations" by van Hasselt et al.
  (https://arxiv.org/abs/1907.03687).

  Args:
    q_tm1: sequence of Q-values at time t-1.
    a_tm1: sequence of action indices at time t-1.
    r_t: sequence of rewards at time t.
    discount_t: sequence of discounts at time t.
    q_t: sequence of Q-values at time t.
    lambda_: mixing parameter lambda, either a scalar (e.g. Peng's Q(lambda)) or
      a sequence (e.g. Watkin's Q(lambda)).
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.
    tx_pair: TxPair of value function transformation and its inverse.

  Returns:
    Q(lambda) temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t, lambda_],
                   [2, 1, 1, 1, 2, {0, 1}])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t, lambda_],
                   [float, int, float, float, float, float])

  qa_tm1 = base.batched_index(q_tm1, a_tm1)
  v_t = jnp.max(q_t, axis=-1)
  target_tm1 = transformed_lambda_returns(
      tx_pair, r_t, discount_t, v_t, lambda_, stop_target_gradients)
  return target_tm1 - qa_tm1


def transformed_retrace(
    q_tm1: Array,
    q_t: Array,
    a_tm1: Array,
    a_t: Array,
    r_t: Array,
    discount_t: Array,
    pi_t: Array,
    mu_t: Array,
    lambda_: float,
    eps: float = 1e-8,
    stop_target_gradients: bool = True,
    tx_pair: TxPair = IDENTITY_PAIR,
) -> Array:
  """Calculates transformed Retrace errors.

  See "Recurrent Experience Replay in Distributed Reinforcement Learning" by
  Kapturowski et al. (https://openreview.net/pdf?id=r1lyTjAqYX).

  Args:
    q_tm1: Q-values at time t-1.
    q_t: Q-values at time t.
    a_tm1: action index at time t-1.
    a_t: action index at time t.
    r_t: reward at time t.
    discount_t: discount at time t.
    pi_t: target policy probs at time t.
    mu_t: behavior policy probs at time t.
    lambda_: scalar mixing parameter lambda.
    eps: small value to add to mu_t for numerical stability.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.
    tx_pair: TxPair of value function transformation and its inverse.

  Returns:
    Transformed Retrace error.
  """
  chex.assert_rank([q_tm1, q_t, a_tm1, a_t, r_t, discount_t, pi_t, mu_t],
                   [2, 2, 1, 1, 1, 1, 2, 1])
  chex.assert_type([q_tm1, q_t, a_tm1, a_t, r_t, discount_t, pi_t, mu_t],
                   [float, float, int, int, float, float, float, float])

  pi_a_t = base.batched_index(pi_t, a_t)
  c_t = jnp.minimum(1.0, pi_a_t / (mu_t + eps)) * lambda_
  target_tm1 = transformed_general_off_policy_returns_from_action_values(
      tx_pair, q_t, a_t, r_t, discount_t, c_t, pi_t, stop_target_gradients)
  q_a_tm1 = base.batched_index(q_tm1, a_tm1)
  return target_tm1 - q_a_tm1


def transformed_n_step_q_learning(
    q_tm1: Array,
    a_tm1: Array,
    target_q_t: Array,
    a_t: Array,
    r_t: Array,
    discount_t: Array,
    n: int,
    stop_target_gradients: bool = True,
    tx_pair: TxPair = IDENTITY_PAIR,
) -> Array:
  """Calculates transformed n-step TD errors.

  See "Recurrent Experience Replay in Distributed Reinforcement Learning" by
  Kapturowski et al. (https://openreview.net/pdf?id=r1lyTjAqYX).

  Args:
    q_tm1: Q-values at times [0, ..., T - 1].
    a_tm1: action index at times [0, ..., T - 1].
    target_q_t: target Q-values at time [1, ..., T].
    a_t: action index at times [[1, ... , T]] used to select target q-values to
      bootstrap from; max(target_q_t) for normal Q-learning, max(q_t) for double
      Q-learning.
    r_t: reward at times [1, ..., T].
    discount_t: discount at times [1, ..., T].
    n: number of steps over which to accumulate reward before bootstrapping.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.
    tx_pair: TxPair of value function transformation and its inverse.

  Returns:
    Transformed N-step TD error.
  """
  chex.assert_rank([q_tm1, target_q_t, a_tm1, a_t, r_t, discount_t],
                   [2, 2, 1, 1, 1, 1])
  chex.assert_type([q_tm1, target_q_t, a_tm1, a_t, r_t, discount_t],
                   [float, float, int, int, float, float])

  v_t = base.batched_index(target_q_t, a_t)
  target_tm1 = transformed_n_step_returns(
      tx_pair, r_t, discount_t, v_t, n,
      stop_target_gradients=stop_target_gradients)
  q_a_tm1 = base.batched_index(q_tm1, a_tm1)
  return target_tm1 - q_a_tm1



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `value_learning.py`."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import distributions
from rlax._src import value_learning


class TDLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.v_tm1 = np.array(
        [1, 1, 1, 1, 1, 1, 1, 1, 1], dtype=np.float32)
    self.r_t = np.array(
        [-1, -1, -1, -1, -1, -1, -1, -1, -1], dtype=np.float32)
    self.discount_t = np.array(
        [0, 0.5, 1, 0, 0.5, 1, 0, 0.5, 1], dtype=np.float32)
    self.v_t = np.array(
        [0, 0, 0, 1, 1, 1, 2, 2, 2], dtype=np.float32)

    self.expected_td = np.array(
        [-2., -2., -2., -2., -1.5, -1., -2., -1., 0.], dtype=np.float32)

  @chex.all_variants()
  def test_td_learning_batch(self):
    """Tests for a full batch."""
    td_learning = self.variant(jax.vmap(value_learning.td_learning))
    # Compute errors in batch.
    actual_td = td_learning(self.v_tm1, self.r_t, self.discount_t, self.v_t)
    # Tets output.
    np.testing.assert_allclose(self.expected_td, actual_td)


class TDLambdaTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.lambda_ = 0.75

    self.v_tm1 = np.array(
        [[1.1, -1.1, 3.1], [2.1, -1.1, -2.1]], dtype=np.float32)
    self.discount_t = np.array(
        [[0., 0.89, 0.85], [0.88, 1., 0.83]], dtype=np.float32)
    self.r_t = np.array(
        [[-1.3, -1.3, 2.3], [1.3, 5.3, -3.3]], dtype=np.float32)
    self.bootstrap_v = np.array([2.2, -1.2], np.float32)

    self.expected = np.array(
        [[-2.4, 3.2732253, 1.0700002],
         [-0.01701999, 2.6529999, -2.196]],
        dtype=np.float32)

  @chex.all_variants()
  def test_batch_compatibility(self):
    """Tests for a full batch."""
    td_lambda = self.variant(jax.vmap(functools.partial(
        value_learning.td_lambda, lambda_=self.lambda_)))
    # Get arguments.
    v_t = np.concatenate([self.v_tm1[:, 1:], self.bootstrap_v[:, None]], axis=1)
    # Test output
    actual = td_lambda(self.v_tm1, self.r_t, self.discount_t, v_t)
    np.testing.assert_allclose(self.expected, actual, rtol=1e-4)


class SarsaTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.q_tm1 = np.array([[1, 1, 0], [1, 1, 0]], dtype=np.float32)
    self.a_tm1 = np.array([0, 1], dtype=np.int32)
    self.r_t = np.array([1, 1], dtype=np.float32)
    self.discount_t = np.array([0, 1], dtype=np.float32)
    self.q_t = np.array([[0, 1, 0], [3, 2, 0]], dtype=np.float32)
    self.a_t = np.array([1, 0], dtype=np.int32)

    self.expected = np.array([0., 3.], dtype=np.float32)

  @chex.all_variants()
  def test_sarsa_batch(self):
    """Tests for a full batch."""
    batch_sarsa = self.variant(jax.vmap(value_learning.sarsa))
    # Test outputs.
    actual = batch_sarsa(self.q_tm1, self.a_tm1, self.r_t, self.discount_t,
                         self.q_t, self.a_t)
    np.testing.assert_allclose(self.expected, actual)


class ExpectedSarsaTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.q_tm1 = np.array(
        [[1, 1, 0.5], [1, 1, 3]], dtype=np.float32)
    self.a_tm1 = np.array(
        [0, 1], dtype=np.int32)
    self.r_t = np.array(
        [4, 1], dtype=np.float32)
    self.discount_t = np.array(
        [1, 1], dtype=np.float32)
    self.q_t = np.array(
        [[1.5, 1, 2], [3, 2, 1]], dtype=np.float32)
    self.probs_a_t = np.array(
        [[0.2, 0.5, 0.3], [0.3, 0.4, 0.3]], dtype=np.float32)

    self.expected = np.array(
        [4.4, 2.], dtype=np.float32)

  @chex.all_variants()
  def test_expected_sarsa_batch(self):
    """Tests for a full batch."""
    expected_sarsa = self.variant(jax.vmap(value_learning.expected_sarsa))
    # Test outputs.
    actual = expected_sarsa(self.q_tm1, self.a_tm1, self.r_t, self.discount_t,
                            self.q_t, self.probs_a_t)
    np.testing.assert_allclose(self.expected, actual)


class SarsaLambdaTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.lambda_ = 0.75

    self.q_tm1 = np.array(
        [[[1.1, 2.1], [-1.1, 1.1], [3.1, -3.1]],
         [[2.1, 3.1], [-1.1, 0.1], [-2.1, -1.1]]],
        dtype=np.float32)
    self.a_tm1 = np.array(
        [[0, 1, 0],
         [1, 0, 0]],
        dtype=np.int32)
    self.discount_t = np.array(
        [[0., 0.89, 0.85],
         [0.88, 1., 0.83]],
        dtype=np.float32)
    self.r_t = np.array(
        [[-1.3, -1.3, 2.3],
         [1.3, 5.3, -3.3]],
        dtype=np.float32)
    self.q_t = np.array(
        [[[1.2, 2.2], [-1.2, 0.2], [2.2, -1.2]],
         [[4.2, 2.2], [1.2, 1.2], [-1.2, -2.2]]],
        dtype=np.float32)
    self.a_t = np.array(
        [[1, 0, 1],
         [1, 1, 0]],
        dtype=np.int32)

    self.expected = np.array(
        [[-2.4, -1.8126001, -1.8200002], [0.25347996, 3.4780002, -2.196]],
        dtype=np.float32)

  @chex.all_variants()
  def test_sarsa_lambda_batch(self):
    """Tests for a full batch."""
    sarsa_lambda = self.variant(jax.vmap(functools.partial(
        value_learning.sarsa_lambda, lambda_=self.lambda_)))
    # Test outputs.
    actual = sarsa_lambda(self.q_tm1, self.a_tm1, self.r_t, self.discount_t,
                          self.q_t, self.a_t)
    np.testing.assert_allclose(self.expected, actual, rtol=1e-4)


class QLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.q_tm1 = np.array([[1, 1, 0], [1, 2, 0]], dtype=np.float32)
    self.a_tm1 = np.array([0, 1], dtype=np.int32)
    self.r_t = np.array([1, 1], dtype=np.float32)
    self.discount_t = np.array([0, 1], dtype=np.float32)
    self.q_t = np.array([[0, 1, 0], [1, 2, 0]], dtype=np.float32)

    self.expected = np.array([0., 1.], dtype=np.float32)

  @chex.all_variants()
  def test_q_learning_batch(self):
    """Tests for a full batch."""
    q_learning = self.variant(jax.vmap(value_learning.q_learning))
    # Test outputs.
    actual = q_learning(self.q_tm1, self.a_tm1, self.r_t, self.discount_t,
                        self.q_t)
    np.testing.assert_allclose(self.expected, actual)


class DoubleQLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.q_tm1 = np.array([[1, 1, 0], [1, 2, 0]], dtype=np.float32)
    self.a_tm1 = np.array([0, 1], dtype=np.int32)
    self.r_t = np.array([1, 1], dtype=np.float32)
    self.discount_t = np.array([0, 1], dtype=np.float32)
    self.q_t_value = np.array([[99, 1, 98], [91, 2, 66]], dtype=np.float32)
    self.q_t_selector = np.array([[2, 10, 1], [11, 20, 1]], dtype=np.float32)

    self.expected = np.array([0., 1.], dtype=np.float32)

  @chex.all_variants()
  def test_double_q_learning_batch(self):
    """Tests for a full batch."""
    double_q_learning = self.variant(jax.vmap(value_learning.double_q_learning))
    # Test outputs.
    actual = double_q_learning(self.q_tm1, self.a_tm1, self.r_t,
                               self.discount_t, self.q_t_value,
                               self.q_t_selector)
    np.testing.assert_allclose(self.expected, actual)


class PersistentQLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.action_gap_scale = 0.25

    self.q_tm1 = np.array([[1, 2], [3, 4], [5, 6]], dtype=np.float32)
    self.a_tm1 = np.array([0, 1, 1], dtype=np.int32)
    self.r_t = np.array([3, 2, 7], dtype=np.float32)
    self.discount_t = np.array([0, 1, 0.5], dtype=np.float32)
    self.q_t = np.array([[11, 12], [20, 16], [-8, -4]], dtype=np.float32)

    self.expected = np.array([2., 17., -1.], dtype=np.float32)

  @chex.all_variants()
  def test_persistent_q_learning_batch(self):
    """Tests for a full batch."""
    # Vmap and optionally compile.
    persistent_q_learning = self.variant(jax.vmap(functools.partial(
        value_learning.persistent_q_learning,
        action_gap_scale=self.action_gap_scale)))
    # Test outputs.
    actual = persistent_q_learning(self.q_tm1, self.a_tm1, self.r_t,
                                   self.discount_t, self.q_t)
    np.testing.assert_allclose(self.expected, actual)


class QVLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.q_tm1 = np.array([[1, 1, 0], [1, 2, 0]], dtype=np.float32)
    self.a_tm1 = np.array([0, 1], dtype=np.int32)
    self.r_t = np.array([1, 1], dtype=np.float32)
    self.discount_t = np.array([0, 1], dtype=np.float32)
    self.v_t = np.array([1, 3], dtype=np.float32)

    self.expected = np.array([0., 2.], dtype=np.float32)

  @chex.all_variants()
  def test_qv_learning_batch(self):
    """Tests for a full batch."""
    batch_qv_learning = self.variant(jax.vmap(value_learning.qv_learning))
    # Test outputs.
    actual = batch_qv_learning(self.q_tm1, self.a_tm1, self.r_t,
                               self.discount_t, self.v_t)
    np.testing.assert_allclose(self.expected, actual)


class QVMaxTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.v_tm1 = np.array(
        [1, 1, 1, 1, 1, 1, 1, 1, 1],
        dtype=np.float32)
    self.r_t = np.array(
        [-1, -1, -1, -1, -1, -1, -1, -1, -1],
        dtype=np.float32)
    self.discount_t = np.array(
        [0, 0.5, 1, 0, 0.5, 1, 0, 0.5, 1],
        dtype=np.float32)
    self.q_t = np.array(
        [[0, -1], [-2, 0], [0, -3], [1, 0], [1, 1], [0, 1],
         [1, 2], [2, -2], [2, 2]],
        dtype=np.float32)

    self.expected = np.array(
        [-2., -2., -2., -2., -1.5, -1., -2., -1., 0.],
        dtype=np.float32)

  @chex.all_variants()
  def test_qv_max_batch(self):
    """Tests for a full batch."""
    qv_max = self.variant(jax.vmap(value_learning.qv_max))
    # Test outputs.
    actual = qv_max(self.v_tm1, self.r_t, self.discount_t, self.q_t)
    np.testing.assert_allclose(self.expected, actual)


class QLambdaTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.lambda_ = 0.75

    self.q_tm1 = np.array(
        [[[1.1, 2.1], [-1.1, 1.1], [3.1, -3.1]],
         [[2.1, 3.1], [-1.1, 0.1], [-2.1, -1.1]]],
        dtype=np.float32)
    self.a_tm1 = np.array(
        [[0, 1, 0],
         [1, 0, 0]],
        dtype=np.int32)
    self.discount_t = np.array(
        [[0., 0.89, 0.85],
         [0.88, 1., 0.83]],
        dtype=np.float32)
    self.r_t = np.array(
        [[-1.3, -1.3, 2.3],
         [1.3, 5.3, -3.3]],
        dtype=np.float32)
    self.q_t = np.array(
        [[[1.2, 2.2], [-1.2, 0.2], [2.2, -1.2]],
         [[4.2, 2.2], [1.2, 1.2], [-1.2, -2.2]]],
        dtype=np.float32)

    self.expected = np.array(
        [[-2.4, 0.427975, 1.07],
         [0.69348, 3.478, -2.196]],
        dtype=np.float32)

  @chex.all_variants()
  def test_q_lambda_batch(self):
    """Tests for a full batch."""
    q_lambda = self.variant(jax.vmap(functools.partial(
        value_learning.q_lambda, lambda_=self.lambda_)))
    # Test outputs.
    actual = q_lambda(self.q_tm1, self.a_tm1, self.r_t, self.discount_t,
                      self.q_t)
    np.testing.assert_allclose(self.expected, actual, rtol=1e-5)


class RetraceTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._lambda = 0.9

    self._qs = np.array(
        [[[1.1, 2.1], [-1.1, 1.1], [3.1, -3.1], [-1.2, 0.0]],
         [[2.1, 3.1], [9.5, 0.1], [-2.1, -1.1], [0.1, 7.4]]],
        dtype=np.float32)
    self._targnet_qs = np.array(
        [[[1.2, 2.2], [-1.2, 0.2], [2.2, -1.2], [-2.25, -6.0]],
         [[4.2, 2.2], [1.2, 1.2], [-1.2, -2.2], [1.5, 1.0]]],
        dtype=np.float32)
    self._actions = np.array(
        [[0, 1, 0, 0], [1, 0, 0, 1]],
        dtype=np.int32)
    self._rewards = np.array(
        [[-1.3, -1.3, 2.3, 42.0],
         [1.3, 5.3, -3.3, -5.0]],
        dtype=np.float32)
    self._pcontinues = np.array(
        [[0., 0.89, 0.85, 0.99],
         [0.88, 1., 0.83, 0.95]],
        dtype=np.float32)
    self._target_policy_probs = np.array(
        [[[0.5, 0.5], [0.2, 0.8], [0.6, 0.4], [0.9, 0.1]],
         [[0.1, 0.9], [1.0, 0.0], [0.3, 0.7], [0.7, 0.3]]],
        dtype=np.float32)
    self._behavior_policy_probs = np.array(
        [[0.5, 0.1, 0.9, 0.3], [0.4, 0.6, 1.0, 0.9]],
        dtype=np.float32)
    self._inputs = [
        self._qs, self._targnet_qs, self._actions,
        self._rewards, self._pcontinues,
        self._target_policy_probs, self._behavior_policy_probs]

    self.expected = np.array(
        [[2.8800001, 3.8934109, 4.5942383],
         [3.1121615e-1, 2.0253206e1, 3.1601219e-3]],
        dtype=np.float32)

  @chex.all_variants()
  def test_retrace_batch(self):
    """Tests for a full batch."""
    retrace = self.variant(jax.vmap(functools.partial(
        value_learning.retrace, lambda_=self._lambda)))
    # Test outputs.
    actual_td = retrace(self._qs[:, :-1], self._targnet_qs[:, 1:],
                        self._actions[:, :-1], self._actions[:, 1:],
                        self._rewards[:, :-1], self._pcontinues[:, :-1],
                        self._target_policy_probs[:, 1:],
                        self._behavior_policy_probs[:, 1:])
    actual_loss = 0.5 * np.square(actual_td)
    np.testing.assert_allclose(self.expected, actual_loss, rtol=1e-5)


def _generate_sorted_support(size):
  """Generate a random support vector."""
  support = np.random.normal(-1.0, 1.0, size=size).astype(np.float32)
  return np.sort(support, axis=-1)


def _generate_weights(size):
  """Generates a weight distribution where half of entries are zero."""
  normal = np.random.normal(-1.0, 1.0, size=size).astype(np.float32)
  mask = (np.random.random(size=size) > 0.5).astype(np.float32)
  return normal * mask


class RetraceContinuousTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._lambda = 0.9 * np.ones((2, 2), dtype=np.float32)

    self._qs = np.array([[1.1, 1.1, 3.1, -1.2], [3.1, 9.5, -2.1, 7.4]],
                        dtype=np.float32)
    self._targnet_qs = np.array([[1.2, 0.2, 2.2, -2.25], [2.2, 1.2, -1.2, 1.0]],
                                dtype=np.float32)
    self._exp_q_t = np.array(
        [[-0.08, 0.84000003, -2.625], [1.2, -1.9, 1.3499999]], dtype=np.float32)
    self._rewards = np.array([[-1.3, -1.3, 2.3, 42.0], [1.3, 5.3, -3.3, -5.0]],
                             dtype=np.float32)
    self._pcontinues = np.array(
        [[0., 0.89, 0.85, 0.99], [0.88, 1., 0.83, 0.95]], dtype=np.float32)
    self._log_rhos = np.array([[2.0794415, -0.4054651, 1.0986122],
                               [0.51082563, -1.2039728, -1.0986123]],
                              dtype=np.float32)
    self._inputs = [
        self._qs, self._targnet_qs, self._exp_q_t, self._rewards,
        self._pcontinues, self._log_rhos
    ]

    self.expected = np.array([[2.880000e+00, 5.643226e+00, 4.594238e+00],
                              [4.236919e-02, 2.044066e+01, 3.160141e-03]],
                             dtype=np.float32)

  @chex.all_variants()
  def test_retrace_batch(self):
    """Tests for a full batch."""
    retrace = self.variant(jax.vmap(value_learning.retrace_continuous))
    # Test outputs.
    actual_td = retrace(self._qs[:, :-1], self._targnet_qs[:,
                                                           1:-1], self._exp_q_t,
                        self._rewards[:, :-1], self._pcontinues[:, :-1],
                        self._log_rhos[:, 1:], self._lambda)
    actual_loss = 0.5 * np.square(actual_td)
    np.testing.assert_allclose(self.expected, actual_loss, rtol=1e-5)

  @chex.all_variants()
  def test_retrace_terminal_batch(self):
    """Tests for a full batch with terminal state."""
    is_terminal = np.array([[0., 1.], [0., 1.]], dtype=np.float32)
    lambda_ = (1. - is_terminal) * self._lambda
    expected = np.array([[2.880000e+00, 1.365213e+00, 4.594238e+00],
                         [2.448239e-02, 1.860500e+01, 3.160141e-03]],
                        dtype=np.float32)
    retrace = self.variant(jax.vmap(value_learning.retrace_continuous))

    # Test outputs.
    actual_td = retrace(self._qs[:, :-1], self._targnet_qs[:, 1:-1],
                        self._exp_q_t, self._rewards[:, :-1],
                        self._pcontinues[:, :-1], self._log_rhos[:,
                                                                 1:], lambda_)
    actual_loss = 0.5 * np.square(actual_td)
    np.testing.assert_allclose(expected, actual_loss, rtol=1e-5)


class L2ProjectTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    old_supports = np.arange(-1, 1., 0.25)
    self.old_supports = np.stack([old_supports, old_supports + 1.])
    weights = self.old_supports.copy()
    weights[0, ::2] = 0.
    weights[1, 1::2] = 0.
    self.weights = weights
    new_supports = np.arange(-1, 1., 0.5)
    self.new_supports = np.stack([new_supports, new_supports + 1.])

    self.expected = np.array([[-0.375, -0.5, 0., 0.875], [0., 0.5, 1., 1.5]],
                             dtype=np.float32)

  @chex.all_variants()
  def test_categorical_l2_project_batch(self):
    """Testsfor a full batch."""
    l2_project = self.variant(jax.vmap(functools.partial(
        value_learning.categorical_l2_project)))
    # Compute projection in batch.
    actual = l2_project(self.old_supports, self.weights, self.new_supports)
    # Test outputs.
    np.testing.assert_allclose(actual, self.expected)


class CategoricalTDLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.atoms = np.array([.5, 1., 1.5], dtype=np.float32)

    self.logits_tm1 = np.array(
        [[0, 9, 0], [9, 0, 9], [0, 9, 0], [9, 9, 0], [9, 0, 9]],
        dtype=np.float32)
    self.r_t = np.array(
        [0.5, 0., 0.5, 0.8, -0.1],
        dtype=np.float32)
    self.discount_t = np.array(
        [0.8, 1., 0.8, 0., 1.],
        dtype=np.float32)
    self.logits_t = np.array(
        [[0, 0, 9], [1, 1, 1], [0, 0, 9], [1, 1, 1], [0, 9, 9]],
        dtype=np.float32)

    self.expected = np.array(
        [8.998915, 3.6932087, 8.998915, 0.69320893, 5.1929307],
        dtype=np.float32)

  @chex.all_variants()
  def test_categorical_td_learning_batch(self):
    """Tests for a full batch."""
    # Not using vmap for atoms.
    def fn(v_logits_tm1, r_t, discount_t, v_logits_t):
      return value_learning.categorical_td_learning(
          v_atoms_tm1=self.atoms,
          v_logits_tm1=v_logits_tm1,
          r_t=r_t,
          discount_t=discount_t,
          v_atoms_t=self.atoms,
          v_logits_t=v_logits_t)
    categorical_td_learning = self.variant(jax.vmap(fn))
    # Test outputs.
    actual = categorical_td_learning(
        self.logits_tm1, self.r_t, self.discount_t, self.logits_t)
    np.testing.assert_allclose(self.expected, actual, rtol=1e-4)


class CategoricalQLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.atoms = np.array([.5, 1., 1.5], dtype=np.float32)

    self.q_logits_tm1 = np.array(
        [[[1, 1, 1], [0, 9, 9], [0, 9, 0], [0, 0, 0]],
         [[9, 9, 0], [9, 0, 9], [0, 0, 0], [9, -9, 0]],
         [[1, 1, 1], [0, 9, 9], [0, 0, 0], [0, 9, 0]],
         [[9, 9, 0], [9, 0, 9], [0, 0, 0], [9, -9, 0]],
         [[9, 9, 0], [9, 0, 9], [0, 0, 0], [9, -9, 0]]],
        dtype=np.float32)
    self.q_logits_t = np.array(
        [[[1, 1, 1], [9, 0, 9], [1, 0, 0], [0, 0, 9]],
         [[9, 9, 0], [9, 0, 0], [1, 1, 1], [9, -9, 0]],
         [[1, 1, 1], [9, 0, 9], [0, 0, 9], [1, 0, 0]],
         [[9, 9, 0], [9, 0, 0], [1, 1, 1], [9, -9, 0]],
         [[9, 9, 0], [9, 0, 0], [0, 9, 9], [9, -9, 0]]],
        dtype=np.float32)
    self.a_tm1 = np.array(
        [2, 1, 3, 0, 1],
        dtype=np.int32)
    self.r_t = np.array(
        [0.5, 0., 0.5, 0.8, -0.1],
        dtype=np.float32)
    self.discount_t = np.array(
        [0.8, 1., 0.8, 0., 1.],
        dtype=np.float32)
    self.inputs = (
        self.q_logits_tm1, self.a_tm1, self.r_t,
        self.discount_t, self.q_logits_t)

    self.expected = np.array(
        [8.998915, 3.6932087, 8.998915, 0.69320893, 5.1929307],
        dtype=np.float32)

  @chex.all_variants()
  def test_categorical_q_learning_batch(self):
    """Tests for a full batch."""
    # Not using vmap for atoms.
    def fn(q_logits_tm1, a_tm1, r_t, discount_t, q_logits_t):
      return value_learning.categorical_q_learning(
          q_atoms_tm1=self.atoms,
          q_logits_tm1=q_logits_tm1,
          a_tm1=a_tm1,
          r_t=r_t,
          discount_t=discount_t,
          q_atoms_t=self.atoms,
          q_logits_t=q_logits_t)
    categorical_q_learning = self.variant(jax.vmap(fn))
    # Test outputs.
    actual = categorical_q_learning(*self.inputs)
    np.testing.assert_allclose(self.expected, actual, rtol=1e-4)


class CategoricalDoubleQLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.atoms = np.array([.5, 1., 1.5], dtype=np.float32)

    self.q_logits_tm1 = np.array(
        [[[1, 1, 1], [0, 9, 9], [0, 9, 0], [0, 0, 0]],
         [[9, 9, 0], [9, 0, 9], [0, 0, 0], [9, -9, 0]],
         [[1, 1, 1], [0, 9, 9], [0, 0, 0], [0, 9, 0]],
         [[9, 9, 0], [9, 0, 9], [0, 0, 0], [9, -9, 0]],
         [[9, 9, 0], [9, 0, 9], [0, 0, 0], [9, -9, 0]]],
        dtype=np.float32)
    self.q_logits_t = np.array(
        [[[1, 1, 1], [9, 0, 9], [1, 0, 0], [0, 0, 9]],
         [[9, 9, 0], [9, 0, 0], [1, 1, 1], [9, -9, 0]],
         [[1, 1, 1], [9, 0, 9], [0, 0, 9], [1, 0, 0]],
         [[9, 9, 0], [9, 0, 0], [1, 1, 1], [9, -9, 0]],
         [[9, 9, 0], [9, 0, 0], [0, 9, 9], [9, -9, 0]]],
        dtype=np.float32)
    self.q_t_selector = np.array(
        [[1, 0, 0, 9], [9, 0, 1, 1], [1, 9, 1, 1], [0, 1, 0, 9], [1, 1, 1, 9]],
        dtype=np.float32)
    self.a_tm1 = np.array(
        [2, 1, 3, 0, 1],
        dtype=np.int32)
    self.r_t = np.array(
        [0.5, 0., 0.5, 0.8, -0.1],
        dtype=np.float32)
    self.discount_t = np.array(
        [0.8, 1., 0.8, 0., 1.],
        dtype=np.float32)
    self.inputs = (
        self.q_logits_tm1, self.a_tm1, self.r_t,
        self.discount_t, self.q_logits_t, self.q_t_selector)

    self.expected = np.array(
        [8.998915, 5.192931, 5.400247, 0.693209, 0.693431],
        dtype=np.float32)

  @chex.all_variants()
  def test_categorical_double_q_learning_batch(self):
    """Tests for a full batch."""
    # Not using vmap for atoms.
    def fn(q_logits_tm1, a_tm1, r_t, discount_t, q_logits_t, q_t_selector):
      return value_learning.categorical_double_q_learning(
          q_atoms_tm1=self.atoms,
          q_logits_tm1=q_logits_tm1,
          a_tm1=a_tm1,
          r_t=r_t,
          discount_t=discount_t,
          q_atoms_t=self.atoms,
          q_logits_t=q_logits_t,
          q_t_selector=q_t_selector)
    categorical_double_q_learning = self.variant(jax.vmap(fn))
    # Test outputs.
    actual = categorical_double_q_learning(*self.inputs)
    np.testing.assert_allclose(self.expected, actual, rtol=1e-4)

  @chex.all_variants()
  def test_single_double_q_learning_eq_batch(self):
    """Tests equivalence to categorical_q_learning when q_t_selector == q_t."""
    # Not using vmap for atoms.
    @self.variant
    @jax.vmap
    def batch_categorical_double_q_learning(
        q_logits_tm1, a_tm1, r_t, discount_t, q_logits_t, q_t_selector):
      return value_learning.categorical_double_q_learning(
          q_atoms_tm1=self.atoms,
          q_logits_tm1=q_logits_tm1,
          a_tm1=a_tm1,
          r_t=r_t,
          discount_t=discount_t,
          q_atoms_t=self.atoms,
          q_logits_t=q_logits_t,
          q_t_selector=q_t_selector)

    @self.variant
    @jax.vmap
    def batch_categorical_q_learning(
        q_logits_tm1, a_tm1, r_t, discount_t, q_logits_t):
      return value_learning.categorical_q_learning(
          q_atoms_tm1=self.atoms,
          q_logits_tm1=q_logits_tm1,
          a_tm1=a_tm1,
          r_t=r_t,
          discount_t=discount_t,
          q_atoms_t=self.atoms,
          q_logits_t=q_logits_t)
    # Double Q-learning estimate with q_t_selector=q_t
    distrib = distributions.softmax()
    # Add batch and time dimension to atoms.
    atoms = jnp.expand_dims(jnp.expand_dims(self.atoms, 0), 0)
    q_t_selector = jnp.sum(distrib.probs(self.q_logits_t) * atoms, axis=-1)
    actual = batch_categorical_double_q_learning(
        self.q_logits_tm1, self.a_tm1, self.r_t, self.discount_t,
        self.q_logits_t, q_t_selector)
    # Q-learning estimate.
    expected = batch_categorical_q_learning(
        self.q_logits_tm1, self.a_tm1, self.r_t, self.discount_t,
        self.q_logits_t)
    # Test equivalence.
    np.testing.assert_allclose(expected, actual)


class QuantileRegressionLossTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.dist_src = np.array([[-1., 3.], [-1., 3.]])
    self.tau_src = np.array([[0.2, 0.7], [0., 0.4]])
    self.dist_target = np.array([[-3., 4., 2.], [-3., 4., 2.]])

    # delta = [[ -2  5  3 ], [ -6  1 -1 ]]
    # Huber(2.)-delta = [[  2  8  4 ], [ 10 .5 .5 ]]
    #
    # First batch element:
    # |tau - Id_{d<0}| = [[ .8 .2 .2 ], [ .3 .7 .3 ]]
    # Loss = 1/3 sum( |delta| . |tau - Id_{d<0}| )  = 2.0
    # Huber(2.)-loss = 2.5
    #
    # Second batch element:
    # |tau - Id_{d<0}| = [[ 1. 0. 0. ], [ .6 .4 .6 ]]
    # Loss = 2.2
    # Huber(2.)-loss = 8.5 / 3
    self.expected_loss = {
        0.: np.array([2.0, 2.2]),
        2.: np.array([2.5, 8.5 / 3.])
    }

  @chex.all_variants()
  @parameterized.named_parameters(
      ('nohuber', 0.),
      ('huber', 2.))
  def test_quantile_regression_loss_batch(self, huber_param):
    """Tests for a full batch."""
    loss_fn = value_learning.quantile_regression_loss
    loss_fn = self.variant(jax.vmap(functools.partial(
        loss_fn, huber_param=huber_param)))
    # Compute quantile regression loss.
    actual = loss_fn(self.dist_src, self.tau_src, self.dist_target)
    # Test outputs in batch.
    np.testing.assert_allclose(actual, self.expected_loss[huber_param],
                               rtol=3e-7)


class QuantileLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.dist_q_tm1 = np.array(  # n_batch = 3, n_taus = 2, n_actions = 4
        [[[0, 1, -5, 6], [-1, 3, 0, -2]],
         [[-5, 9, -5, 6], [2, 3, 1, -4]],
         [[5, 1, -5, 7], [-1, 3, 0, -2]]],
        dtype=np.float32)
    self.tau_q_tm1 = np.array(
        [[0.2, 0.7],
         [0.1, 0.5],
         [0.3, 0.4]],
        dtype=np.float32)
    self.a_tm1 = np.array(
        [1, 2, 0],
        dtype=np.int32)
    self.r_t = np.array(
        [0.5, -1., 0.],
        dtype=np.float32)
    self.discount_t = np.array(
        [0.5, 0., 1],
        dtype=np.float32)
    self.dist_q_t = np.array(
        [[[0, 5, 2, 2], [0, -3, 2, 2]],
         [[-3, -1, 4, -3], [1, 3, 1, -4]],
         [[-2, 2, -5, -7], [1, 3, 2, -2]]],
        dtype=np.float32)
    self.dist_q_t_selector = np.array(
        [[[0, 7, 2, -2], [0, 4, 2, 2]],
         [[-3, -1, 4, 3], [1, 3, 1, 4]],
         [[-1, -2, -5, -6], [-1, -5, 2, -2]]],
        dtype=np.float32)

    # Scenario 1: these match the dist_q_t_selector above
    self.greedy_probs_a_t = np.array(
        [[0, 1, 0, 0],
         [0, 0, 0, 1],
         [1, 0, 0, 0]],
        dtype=np.float32)

    dist_qa_tm1 = np.array(
        [[1, 3], [-5, 1], [5, -1]],
        dtype=np.float32)
    # dist_qa_tm1                                      [ 1,  3]
    #     (batch x n_tau)                          =   [-5,  1]
    #                                                  [ 5, -1]
    # dist_q_t_selector[mean]                          [ 0.0,  5.5,  2.0,  0.0]
    #     (batch x n_actions)                      =   [-1.0,  1.0,  2.5,  3.5]
    #                                                  [-1.0, -3.5, -1.5, -4.0]
    # a_t = argmax_a dist_q_t_selector                 [1]
    #     (batch)                                  =   [3]
    #                                                  [0]
    # dist_qa_t                                        [ 5, -3]
    #     (batch x n_taus)                         =   [-3, -4]
    #                                                  [-2,  1]
    # target = r + gamma * dist_qa_t                   [ 3, -1]
    #     (batch x n_taus)                         =   [-1, -1]
    #                                                  [-2,  1]
    dist_target = np.array(
        [[3, -1], [-1, -1], [-2, 1]],
        dtype=np.float32)

    # Use qr loss to compute expected results (itself tested explicitly in
    # distributions_test.py).
    self.expected = {}
    for huber_param in [0.0, 1.0]:
      self.expected[huber_param] = np.array(  # loop over batch
          [value_learning.quantile_regression_loss(dqa, tau, dt, huber_param)
           for (dqa, tau, dt) in zip(dist_qa_tm1, self.tau_q_tm1, dist_target)],
          dtype=np.float32)

    # Scenario 2:
    # bootstrap targets are not an argmax, but averaging across actions
    self.uniform_probs_a_t = np.ones((3, 4), dtype=np.float32) / 4.

    # all_action_targets = r + gamma * dist_q_t  (batch x n_taus x n_actions)
    dts = np.array(
        [[[0.5, 3, 1.5, 1.5], [0.5, -1, 1.5, 1.5]],
         [[-1, -1, -1, -1], [-1, -1, -1, -1]],
         [[-2, 2, -5, -7], [1, 3, 2, -2]]],
        dtype=np.float32)

    self.uniform_expected = {}
    for huber_param in [0.0, 1.0]:
      all_action_losses = [  # loop over actions and batch
          [value_learning.quantile_regression_loss(dqa, tau, dt, huber_param)
           for (dqa, tau, dt) in zip(dist_qa_tm1, self.tau_q_tm1, dts[:, :, a])]
          for a in range(4)]
      # uniform probabilities means the loss is averaged over actions
      self.uniform_expected[huber_param] = np.array(
          all_action_losses, dtype=np.float32).mean(axis=0)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('nohuber', 0.0),
      ('huber', 1.0))
  def test_quantile_q_learning_batch(self, huber_param):
    """Tests for a full batch."""
    quantile_q_learning = self.variant(jax.vmap(functools.partial(
        value_learning.quantile_q_learning, huber_param=huber_param)))
    # Test outputs.
    actual = quantile_q_learning(
        self.dist_q_tm1, self.tau_q_tm1, self.a_tm1, self.r_t, self.discount_t,
        self.dist_q_t_selector, self.dist_q_t)
    np.testing.assert_allclose(self.expected[huber_param], actual, rtol=1e-5)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('nohuber', 0.0),
      ('huber', 1.0))
  def test_quantile_expected_sarsa_batch_greedy(self, huber_param):
    """Tests for a full batch."""
    quantile_expected_sarsa = self.variant(jax.vmap(functools.partial(
        value_learning.quantile_expected_sarsa, huber_param=huber_param)))
    # Test outputs.
    actual = quantile_expected_sarsa(
        self.dist_q_tm1, self.tau_q_tm1, self.a_tm1, self.r_t, self.discount_t,
        self.dist_q_t, self.greedy_probs_a_t)
    np.testing.assert_allclose(self.expected[huber_param], actual, rtol=1e-5)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('nohuber', 0.0),
      ('huber', 1.0))
  def test_quantile_expected_sarsa_batch_uniform(self, huber_param):
    """Tests for a full batch."""
    quantile_expected_sarsa = self.variant(jax.vmap(functools.partial(
        value_learning.quantile_expected_sarsa, huber_param=huber_param)))
    # Test outputs.
    actual = quantile_expected_sarsa(
        self.dist_q_tm1, self.tau_q_tm1, self.a_tm1, self.r_t, self.discount_t,
        self.dist_q_t, self.uniform_probs_a_t)
    np.testing.assert_allclose(
        self.uniform_expected[huber_param], actual, rtol=1e-5)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""JAX functions for V-Trace algorithm.

V-Trace is a form of importance sampling correction that was introduced by
Espeholt et al. in the context of an off-policy actor-critic agent (IMPALA).
This subpackage implements the specific targets used in IMPALA to implement
both the value and the policy. Note however that the V-Trace return estimate is
a special case of the multistep return estimates from `multistep.py`.
"""

import collections

import chex
import jax
import jax.numpy as jnp


Array = chex.Array
Numeric = chex.Numeric
VTraceOutput = collections.namedtuple(
    'vtrace_output', ['errors', 'pg_advantage', 'q_estimate'])


def vtrace(
    v_tm1: Array,
    v_t: Array,
    r_t: Array,
    discount_t: Array,
    rho_tm1: Array,
    lambda_: Numeric = 1.0,
    clip_rho_threshold: float = 1.0,
    stop_target_gradients: bool = True,
) -> Array:
  """Calculates V-Trace errors from importance weights.

  V-trace computes TD-errors from multistep trajectories by applying
  off-policy corrections based on clipped importance sampling ratios.

  See "IMPALA: Scalable Distributed Deep-RL with Importance Weighted Actor
  Learner Architectures" by Espeholt et al. (https://arxiv.org/abs/1802.01561).

  Args:
    v_tm1: values at time t-1.
    v_t: values at time t.
    r_t: reward at time t.
    discount_t: discount at time t.
    rho_tm1: importance sampling ratios at time t-1.
    lambda_: mixing parameter; a scalar or a vector for timesteps t.
    clip_rho_threshold: clip threshold for importance weights.
    stop_target_gradients: whether or not to apply stop gradient to targets.

  Returns:
    V-Trace error.
  """
  chex.assert_rank([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [1, 1, 1, 1, 1, {0, 1}])
  chex.assert_type([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [float, float, float, float, float, float])
  chex.assert_equal_shape([v_tm1, v_t, r_t, discount_t, rho_tm1])

  # Clip importance sampling ratios.
  c_tm1 = jnp.minimum(1.0, rho_tm1) * lambda_
  clipped_rhos_tm1 = jnp.minimum(clip_rho_threshold, rho_tm1)

  # Compute the temporal difference errors.
  td_errors = clipped_rhos_tm1 * (r_t + discount_t * v_t - v_tm1)

  # Work backwards computing the td-errors.
  def _body(acc, xs):
    td_error, discount, c = xs
    acc = td_error + discount * c * acc
    return acc, acc

  _, errors = jax.lax.scan(
      _body, 0.0, (td_errors, discount_t, c_tm1), reverse=True)

  # Return errors, maybe disabling gradient flow through bootstrap targets.
  return jax.lax.select(
      stop_target_gradients,
      jax.lax.stop_gradient(errors + v_tm1) - v_tm1,
      errors)


def leaky_vtrace(
    v_tm1: Array,
    v_t: Array,
    r_t: Array,
    discount_t: Array,
    rho_tm1: Array,
    alpha_: float = 1.0,
    lambda_: Numeric = 1.0,
    clip_rho_threshold: float = 1.0,
    stop_target_gradients: bool = True):
  """Calculates Leaky V-Trace errors from importance weights.

  Leaky-Vtrace is a combination of Importance sampling and V-trace, where the
  degree of mixing is controlled by a scalar `alpha` (that may be meta-learnt).

  See "Self-Tuning Deep Reinforcement Learning"
  by Zahavy et al. (https://arxiv.org/abs/2002.12928)

  Args:
    v_tm1: values at time t-1.
    v_t: values at time t.
    r_t: reward at time t.
    discount_t: discount at time t.
    rho_tm1: importance weights at time t-1.
    alpha_: mixing parameter for Importance Sampling and V-trace.
    lambda_: mixing parameter; a scalar or a vector for timesteps t.
    clip_rho_threshold: clip threshold for importance weights.
    stop_target_gradients: whether or not to apply stop gradient to targets.

  Returns:
    Leaky V-Trace error.
  """
  chex.assert_rank([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [1, 1, 1, 1, 1, {0, 1}])
  chex.assert_type([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [float, float, float, float, float, float])
  chex.assert_equal_shape([v_tm1, v_t, r_t, discount_t, rho_tm1])

  # Mix clipped and unclipped importance sampling ratios.
  c_tm1 = (
      (1 - alpha_) * rho_tm1 + alpha_ * jnp.minimum(1.0, rho_tm1)) * lambda_
  clipped_rhos_tm1 = (
      (1 - alpha_) * rho_tm1 + alpha_ * jnp.minimum(clip_rho_threshold, rho_tm1)
  )

  # Compute the temporal difference errors.
  td_errors = clipped_rhos_tm1 * (r_t + discount_t * v_t - v_tm1)

  # Work backwards computing the td-errors.
  def _body(acc, xs):
    td_error, discount, c = xs
    acc = td_error + discount * c * acc
    return acc, acc

  _, errors = jax.lax.scan(
      _body, 0.0, (td_errors, discount_t, c_tm1), reverse=True)

  # Return errors, maybe disabling gradient flow through bootstrap targets.
  return jax.lax.select(
      stop_target_gradients,
      jax.lax.stop_gradient(errors + v_tm1) - v_tm1,
      errors)


def vtrace_td_error_and_advantage(
    v_tm1: Array,
    v_t: Array,
    r_t: Array,
    discount_t: Array,
    rho_tm1: Array,
    lambda_: Numeric = 1.0,
    clip_rho_threshold: float = 1.0,
    clip_pg_rho_threshold: float = 1.0,
    stop_target_gradients: bool = True,
) -> VTraceOutput:
  """Calculates V-Trace errors and PG advantage from importance weights.

  This functions computes the TD-errors and policy gradient Advantage terms
  as used by the IMPALA distributed actor-critic agent.

  See "IMPALA: Scalable Distributed Deep-RL with Importance Weighted Actor
  Learner Architectures" by Espeholt et al. (https://arxiv.org/abs/1802.01561)

  Args:
    v_tm1: values at time t-1.
    v_t: values at time t.
    r_t: reward at time t.
    discount_t: discount at time t.
    rho_tm1: importance weights at time t-1.
    lambda_: mixing parameter; a scalar or a vector for timesteps t.
    clip_rho_threshold: clip threshold for importance ratios.
    clip_pg_rho_threshold: clip threshold for policy gradient importance ratios.
    stop_target_gradients: whether or not to apply stop gradient to targets.

  Returns:
    a tuple of V-Trace error, policy gradient advantage, and estimated Q-values.
  """
  chex.assert_rank([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [1, 1, 1, 1, 1, {0, 1}])
  chex.assert_type([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [float, float, float, float, float, float])
  chex.assert_equal_shape([v_tm1, v_t, r_t, discount_t, rho_tm1])

  # If scalar make into vector.
  lambda_ = jnp.ones_like(discount_t) * lambda_

  errors = vtrace(
      v_tm1, v_t, r_t, discount_t, rho_tm1,
      lambda_, clip_rho_threshold, stop_target_gradients)
  targets_tm1 = errors + v_tm1
  q_bootstrap = jnp.concatenate([
      lambda_[:-1] * targets_tm1[1:] + (1 - lambda_[:-1]) * v_tm1[1:],
      v_t[-1:],
  ], axis=0)
  q_estimate = r_t + discount_t * q_bootstrap
  clipped_pg_rho_tm1 = jnp.minimum(clip_pg_rho_threshold, rho_tm1)
  pg_advantages = clipped_pg_rho_tm1 * (q_estimate - v_tm1)
  return VTraceOutput(
      errors=errors, pg_advantage=pg_advantages, q_estimate=q_estimate)


def leaky_vtrace_td_error_and_advantage(
    v_tm1: chex.Array,
    v_t: chex.Array,
    r_t: chex.Array,
    discount_t: chex.Array,
    rho_tm1: chex.Array,
    alpha: float = 1.0,
    lambda_: Numeric = 1.0,
    clip_rho_threshold: float = 1.0,
    clip_pg_rho_threshold: float = 1.0,
    stop_target_gradients: bool = True,
) -> VTraceOutput:
  """Calculates Leaky V-Trace errors and PG advantage from importance weights.

  This functions computes the Leaky V-Trace TD-errors and policy gradient
  Advantage terms as used by the IMPALA distributed actor-critic agent.

  Leaky-Vtrace is a combination of Importance sampling and V-trace, where the
  degree of mixing is controlled by a scalar `alpha` (that may be meta-learnt).

  See "Self-Tuning Deep Reinforcement Learning"
  by Zahavy et al. (https://arxiv.org/abs/2002.12928) and
  "IMPALA: Scalable Distributed Deep-RL with Importance Weighted Actor
  Learner Architectures" by Espeholt et al. (https://arxiv.org/abs/1802.01561)

  Args:
    v_tm1: values at time t-1.
    v_t: values at time t.
    r_t: reward at time t.
    discount_t: discount at time t.
    rho_tm1: importance weights at time t-1.
    alpha: mixing the clipped importance sampling weights with unclipped ones.
    lambda_: mixing parameter; a scalar or a vector for timesteps t.
    clip_rho_threshold: clip threshold for importance ratios.
    clip_pg_rho_threshold: clip threshold for policy gradient importance ratios.
    stop_target_gradients: whether or not to apply stop gradient to targets.

  Returns:
    a tuple of V-Trace error, policy gradient advantage, and estimated Q-values.
  """
  chex.assert_rank([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [1, 1, 1, 1, 1, {0, 1}])
  chex.assert_type([v_tm1, v_t, r_t, discount_t, rho_tm1, lambda_],
                   [float, float, float, float, float, float])
  chex.assert_equal_shape([v_tm1, v_t, r_t, discount_t, rho_tm1])

  # If scalar make into vector.
  lambda_ = jnp.ones_like(discount_t) * lambda_

  errors = leaky_vtrace(
      v_tm1, v_t, r_t, discount_t, rho_tm1, alpha,
      lambda_, clip_rho_threshold, stop_target_gradients)
  targets_tm1 = errors + v_tm1
  q_bootstrap = jnp.concatenate([
      lambda_[:-1] * targets_tm1[1:] + (1 - lambda_[:-1]) * v_tm1[1:],
      v_t[-1:],
  ], axis=0)
  q_estimate = r_t + discount_t * q_bootstrap
  clipped_pg_rho_tm1 = ((1 - alpha) * rho_tm1 + alpha *
                        jnp.minimum(clip_pg_rho_threshold, rho_tm1))
  pg_advantages = clipped_pg_rho_tm1 * (q_estimate - v_tm1)
  return VTraceOutput(
      errors=errors, pg_advantage=pg_advantages, q_estimate=q_estimate)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `multistep.py`."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import multistep


class LambdaReturnsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.lambda_ = 0.75

    self.r_t = np.array(
        [[1.0, 0.0, -1.0, 0.0, 1.0], [0.5, 0.8, -0.7, 0.0, 2.1]])
    self.discount_t = np.array(
        [[0.5, 0.9, 1.0, 0.5, 0.8], [0.9, 0.5, 0.3, 0.8, 0.7]])
    self.v_t = np.array(
        [[3.0, 1.0, 5.0, -5.0, 3.0], [-1.7, 1.2, 2.3, 2.2, 2.7]])

    self.expected = np.array(
        [[1.6460547, 0.72281253, 0.7375001, 0.6500001, 3.4],
         [0.7866317, 0.9913063, 0.1101501, 2.834, 3.99]],
        dtype=np.float32)

  @chex.all_variants()
  def test_lambda_returns_batch(self):
    """Tests for a full batch."""
    lambda_returns = self.variant(jax.vmap(functools.partial(
        multistep.lambda_returns, lambda_=self.lambda_)))
    # Compute lambda return in batch.
    actual = lambda_returns(self.r_t, self.discount_t, self.v_t)
    # Test return estimate.
    np.testing.assert_allclose(self.expected, actual, rtol=1e-5)


class DiscountedReturnsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.r_t = np.array(
        [[1.0, 0.0, -1.0, 0.0, 1.0], [0.5, 0.8, -0.7, 0.0, 2.1]])
    self.discount_t = np.array(
        [[0.5, 0.9, 1.0, 0.5, 0.8], [0.9, 0.5, 0.3, 0.8, 0.7]])
    self.v_t = np.array(
        [[3.0, 1.0, 5.0, -5.0, 3.0], [-1.7, 1.2, 2.3, 2.2, 2.7]])
    self.bootstrap_v = np.array([v[-1] for v in self.v_t])

    self.expected = np.array(
        [[1.315, 0.63000005, 0.70000005, 1.7, 3.4],
         [1.33592, 0.9288, 0.2576, 3.192, 3.9899998]],
        dtype=np.float32)

  @chex.all_variants()
  def test_discounted_returns_batch(self):
    """Tests for a single element."""
    discounted_returns = self.variant(jax.vmap(multistep.discounted_returns))
    # Compute discounted return.
    actual_scalar = discounted_returns(self.r_t, self.discount_t,
                                       self.bootstrap_v)
    actual_vector = discounted_returns(self.r_t, self.discount_t, self.v_t)
    # Test output.
    np.testing.assert_allclose(self.expected, actual_scalar, rtol=1e-5)
    np.testing.assert_allclose(self.expected, actual_vector, rtol=1e-5)


class NStepBootstrappedReturnsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.r_t = np.array(
        [[1.0, 0.0, -1.0, 0.0, 1.0], [0.5, 0.8, -0.7, 0.0, 2.1]])
    self.discount_t = np.array(
        [[0.5, 0.9, 1.0, 0.5, 0.8], [0.9, 0.5, 0.3, 0.8, 0.7]])
    self.v_t = np.array(
        [[3.0, 1.0, 5.0, -5.0, 3.0], [-1.7, 1.2, 2.3, 2.2, 2.7]])

    # Different expected results for different values of n.
    self.expected = {}
    self.expected[3] = np.array(
        [[2.8, -3.15, 0.7, 1.7, 3.4], [1.2155, 0.714, 0.2576, 3.192, 3.99]],
        dtype=np.float32)
    self.expected[5] = np.array(
        [[1.315, 0.63, 0.7, 1.7, 3.4], [1.33592, 0.9288, 0.2576, 3.192, 3.99]],
        dtype=np.float32)
    self.expected[7] = np.array(
        [[1.315, 0.63, 0.7, 1.7, 3.4], [1.33592, 0.9288, 0.2576, 3.192, 3.99]],
        dtype=np.float32)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('smaller_n', 3,), ('equal_n', 5,), ('bigger_n', 7,))
  def test_n_step_sequence_returns_batch(self, n):
    """Tests for a full batch."""
    n_step_returns = self.variant(jax.vmap(functools.partial(
        multistep.n_step_bootstrapped_returns, n=n)))
    # Compute n-step return in batch.
    actual = n_step_returns(self.r_t, self.discount_t, self.v_t)
    # Test return estimate.
    np.testing.assert_allclose(self.expected[n], actual, rtol=1e-5)

  def test_reduces_to_lambda_returns(self):
    """Test function is the same as lambda_returns when n is sequence length."""
    lambda_t = 0.75
    n = len(self.r_t[0])
    expected = multistep.lambda_returns(self.r_t[0], self.discount_t[0],
                                        self.v_t[0], lambda_t)
    actual = multistep.n_step_bootstrapped_returns(self.r_t[0],
                                                   self.discount_t[0],
                                                   self.v_t[0], n, lambda_t)
    np.testing.assert_allclose(expected, actual, rtol=1e-5)


class TDErrorTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.r_t = np.array(
        [[1.0, 0.0, -1.0, 0.0, 1.0], [0.5, 0.8, -0.7, 0.0, 2.1]])
    self.discount_t = np.array(
        [[0.5, 0.9, 1.0, 0.5, 0.8], [0.9, 0.5, 0.3, 0.8, 0.7]])
    self.rho_tm1 = np.array(
        [[0.5, 0.9, 1.3, 0.2, 0.8], [2., 0.1, 1., 0.4, 1.7]])
    self.values = np.array(
        [[3.0, 1.0, 5.0, -5.0, 3.0, 1.], [-1.7, 1.2, 2.3, 2.2, 2.7, 2.]])

  @chex.all_variants()
  def test_importance_corrected_td_errors_batch(self):
    """Tests equivalence to computing the error from a the lambda-return."""
    # Vmap and optionally compile.
    lambda_returns = self.variant(jax.vmap(multistep.lambda_returns))
    td_errors = self.variant(jax.vmap(multistep.importance_corrected_td_errors))
    # Compute multistep td-error with recursion on deltas.
    td_direct = td_errors(self.r_t, self.discount_t, self.rho_tm1,
                          np.ones_like(self.discount_t), self.values)
    # Compute off-policy corrected return, and derive td-error from it.
    ls_ = np.concatenate((self.rho_tm1[:, 1:], [[1.], [1.]]), axis=1)
    td_from_returns = self.rho_tm1 * (
        lambda_returns(self.r_t, self.discount_t, self.values[:, 1:], ls_) -
        self.values[:, :-1])
    # Check equivalence.
    np.testing.assert_allclose(td_direct, td_from_returns, rtol=1e-5)


class TruncatedGeneralizedAdvantageEstimationTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.r_t = jnp.array([[0., 0., 1., 0., -0.5],
                          [0., 0., 0., 0., 1.]])
    self.v_t = jnp.array([[1., 4., -3., -2., -1., -1.],
                          [-3., -2., -1., 0.0, 5., -1.]])
    self.discount_t = jnp.array([[0.99, 0.99, 0.99, 0.99, 0.99],
                                 [0.9, 0.9, 0.9, 0.0, 0.9]])
    self.dummy_rho_tm1 = jnp.array([[1., 1., 1., 1., 1],
                                    [1., 1., 1., 1., 1.]])
    self.array_lambda = jnp.array([[0.9, 0.9, 0.9, 0.9, 0.9],
                                   [0.9, 0.9, 0.9, 0.9, 0.9]])

    # Different expected results for different values of lambda.
    self.expected = {}
    self.expected[1.] = np.array(
        [[-1.45118, -4.4557, 2.5396, 0.5249, -0.49],
         [3., 2., 1., 0., -4.9]],
        dtype=np.float32)
    self.expected[0.7] = np.array(
        [[-0.676979, -5.248167, 2.4846, 0.6704, -0.49],
         [2.2899, 1.73, 1., 0., -4.9]],
        dtype=np.float32)
    self.expected[0.4] = np.array(
        [[0.56731, -6.042, 2.3431, 0.815, -0.49],
         [1.725, 1.46, 1., 0., -4.9]],
        dtype=np.float32)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('lambda1', 1.0),
      ('lambda0.7', 0.7),
      ('lambda0.4', 0.4))
  def test_truncated_gae(self, lambda_):
    """Tests truncated GAE for a full batch."""
    batched_advantage_fn_variant = self.variant(jax.vmap(
        multistep.truncated_generalized_advantage_estimation,
        in_axes=(0, 0, None, 0), out_axes=0))
    actual = batched_advantage_fn_variant(
        self.r_t, self.discount_t, lambda_, self.v_t)
    np.testing.assert_allclose(self.expected[lambda_], actual, atol=1e-3)

  @chex.all_variants()
  def test_array_lambda(self):
    """Tests that truncated GAE is consistent with scalar or array lambda_."""
    scalar_lambda_fn = self.variant(jax.vmap(
        multistep.truncated_generalized_advantage_estimation,
        in_axes=(0, 0, None, 0), out_axes=0))
    array_lambda_fn = self.variant(jax.vmap(
        multistep.truncated_generalized_advantage_estimation))
    scalar_lambda_result = scalar_lambda_fn(
        self.r_t, self.discount_t, 0.9, self.v_t)
    array_lambda_result = array_lambda_fn(
        self.r_t, self.discount_t, self.array_lambda, self.v_t)
    np.testing.assert_allclose(scalar_lambda_result, array_lambda_result,
                               atol=1e-3)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('lambda_1', 1.0),
      ('lambda_0.7', 0.7),
      ('lambda_0.4', 0.4))
  def test_gae_as_special_case_of_importance_corrected_td_errors(self, lambda_):
    """Tests truncated GAE.

    Tests that truncated GAE yields same output as importance corrected
    td errors with dummy ratios.

    Args:
      lambda_: a lambda to use in GAE.
    """
    batched_gae_fn_variant = self.variant(jax.vmap(
        multistep.truncated_generalized_advantage_estimation,
        in_axes=(0, 0, None, 0), out_axes=0))
    gae_result = batched_gae_fn_variant(
        self.r_t, self.discount_t, lambda_, self.v_t)

    batched_ictd_errors_fn_variant = self.variant(jax.vmap(
        multistep.importance_corrected_td_errors))
    ictd_errors_result = batched_ictd_errors_fn_variant(
        self.r_t,
        self.discount_t,
        self.dummy_rho_tm1,
        jnp.ones_like(self.discount_t) * lambda_,
        self.v_t)
    np.testing.assert_allclose(gae_result, ictd_errors_result, atol=1e-3)

if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for creating and managing moving averages."""

import chex
import jax
import jax.numpy as jnp


@chex.dataclass(frozen=True)
class EmaMoments:
  """data-class holding the latest mean and variance estimates."""
  # The tree of means.
  mean: chex.ArrayTree
  # The tree of variances.
  variance: chex.ArrayTree


@chex.dataclass(frozen=True)
class EmaState:
  """data-class holding the exponential moving average state."""
  # The tree of exponential moving averages of the values
  mu: chex.ArrayTree
  # The tree of exponential moving averages of the squared values
  nu: chex.ArrayTree
  # The product of the all decays from the start of accumulating.
  decay_product: float

  def debiased_moments(self):
    """Returns debiased moments as in Adam."""
    tiny = jnp.finfo(self.decay_product).tiny
    debias = 1.0 / jnp.maximum(1 - self.decay_product, tiny)
    mean = jax.tree_map(lambda m1: m1 * debias, self.mu)
    # This computation of the variance may lose some numerical precision, if
    # the mean is not approximately zero.
    variance = jax.tree_map(
        lambda m2, m: jnp.maximum(0.0, m2 * debias - jnp.square(m)),
        self.nu, mean)
    return EmaMoments(mean=mean, variance=variance)


def create_ema(decay=0.999, pmean_axis_name=None):
  """An updater of moments.

  Given a `tree` it will track first and second moments of the leaves.

  Args:
    decay: The decay of the moments. I.e., the learning rate is `1 - decay`.
    pmean_axis_name: If not None, use lax.pmean to average the moment updates.

  Returns:
    Two functions: `(init_state, update_moments)`.
  """

  def init_state(template_tree):
    zeros = jax.tree_map(lambda x: jnp.zeros_like(jnp.mean(x)), template_tree)
    scalar_zero = jnp.ones([], dtype=jnp.float32)
    return EmaState(mu=zeros, nu=zeros, decay_product=scalar_zero)

  def _update(moment, value):
    mean = jnp.mean(value)
    # Compute the mean across all learner devices involved in the `pmap`.
    if pmean_axis_name is not None:
      mean = jax.lax.pmean(mean, axis_name=pmean_axis_name)
    return decay * moment + (1 - decay) * mean

  def update_moments(tree, state):
    squared_tree = jax.tree_map(jnp.square, tree)
    mu = jax.tree_map(_update, state.mu, tree)
    nu = jax.tree_map(_update, state.nu, squared_tree)
    state = EmaState(
        mu=mu, nu=nu, decay_product=state.decay_product * decay)
    return state.debiased_moments(), state

  return init_state, update_moments



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `base.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import base


class OneHotTest(parameterized.TestCase):

  def test_one_hot(self):
    num_classes = 3
    indices = jnp.array(
        [[[1., 2., 3.], [1., 2., 2.]]])
    expected_result = jnp.array([
        [[[0., 1., 0.], [0., 0., 1.], [0., 0., 0.]],
         [[0., 1., 0.], [0., 0., 1.], [0., 0., 1.]]]])
    result = base.one_hot(indices, num_classes)
    np.testing.assert_array_almost_equal(result, expected_result)


class BroadcastTest(parameterized.TestCase):

  @parameterized.parameters(
      ([1], [1, 2, 3], [1, 1, 1]),
      ([1, 2, 1], [1, 2, 3], [1, 2, 1]),
      ([2, 1, 2], [2, 2, 2, 3], [2, 1, 2, 1]),
      ([1, 2, 4], [1, 2, 4], [1, 2, 4]),
  )
  def test_lhs_broadcasting(
      self, source_shape, target_shape, expected_result_shape):
    source = jnp.ones(shape=source_shape, dtype=jnp.float32)
    target = jnp.ones(shape=target_shape, dtype=jnp.float32)
    expected_result = jnp.ones(shape=expected_result_shape, dtype=jnp.float32)
    result = base.lhs_broadcast(source, target)
    np.testing.assert_array_almost_equal(result, expected_result)

  def test_lhs_broadcast_raises(self):
    source = jnp.ones(shape=(1, 2), dtype=jnp.float32)
    target = jnp.ones(shape=(1, 3, 1, 1), dtype=jnp.float32)
    with self.assertRaisesRegex(ValueError, 'source shape'):
      base.lhs_broadcast(source, target)


class ReplaceTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.data = jnp.array([
        [1, 2, 3, 4, 5, 6],
        [-1, -2, -3, -4, -5, -6]
    ])
    self.replacement = self.data * 10
    self.mask = jnp.array([0, 1])

  def test_replace_masked(self):
    output = base.replace_masked(self.data, self.replacement, self.mask)
    expected_output = jnp.array([
        [1, 2, 3, 4, 5, 6],
        [-10, -20, -30, -40, -50, -60],
    ])
    # Test output.
    np.testing.assert_allclose(output, expected_output)

  def test_replace_masked_zeros(self):
    output = base.replace_masked(self.data, None, self.mask)
    expected_output = jnp.array([
        [1, 2, 3, 4, 5, 6],
        [0, 0, 0, 0, 0, 0],
    ])
    # Test output.
    np.testing.assert_allclose(output, expected_output)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX implementation of common losses.

Deep reinforcement learning algorithms are often expressed as gradients of
suitable pseudo-loss functions constructed from the observations and rewards
collected in the environment. In this subpackage we collate common mathematical
transformations used to construct such losses.
"""

import functools

from typing import Optional, Union
import chex
import jax
import jax.numpy as jnp
from rlax._src import general_value_functions
from rlax._src import value_learning

Scalar = chex.Scalar
Array = chex.Array


def l2_loss(predictions: Array,
            targets: Optional[Array] = None) -> Array:
  """Caculates the L2 loss of predictions wrt targets.

  If targets are not provided this function acts as an L2-regularizer for preds.

  Note: the 0.5 term is standard in "Pattern Recognition and Machine Learning"
  by Bishop, but not "The Elements of Statistical Learning" by Tibshirani.

  Args:
    predictions: a vector of arbitrary shape.
    targets: a vector of shape compatible with predictions.

  Returns:
    a vector of same shape of `predictions`.
  """
  if targets is None:
    targets = jnp.zeros_like(predictions)
  chex.assert_type([predictions, targets], float)
  return 0.5 * (predictions - targets)**2


def likelihood(predictions: Array, targets: Array) -> Array:
  """Calculates the likelihood of predictions wrt targets.

  Args:
    predictions: a vector of arbitrary shape.
    targets: a vector of shape compatible with predictions.

  Returns:
    a vector of same shape of `predictions`.
  """
  chex.assert_type([predictions, targets], float)
  likelihood_vals = predictions**targets * (1. - predictions)**(1. - targets)
  # Note: 0**0 evaluates to NaN on TPUs, manually set these cases to 1.
  filter_indices = jnp.logical_or(
      jnp.logical_and(targets == 1, predictions == 1),
      jnp.logical_and(targets == 0, predictions == 0))
  return jnp.where(filter_indices, 1, likelihood_vals)


def log_loss(
    predictions: Array,
    targets: Array,
) -> Array:
  """Calculates the log loss of predictions wrt targets.

  Args:
    predictions: a vector of probabilities of arbitrary shape.
    targets: a vector of probabilities of shape compatible with predictions.

  Returns:
    a vector of same shape of `predictions`.
  """
  chex.assert_type([predictions, targets], float)
  return -jnp.log(likelihood(predictions, targets))


def pixel_control_loss(
    observations: Array,
    actions: Array,
    action_values: Array,
    discount_factor: Union[Array, Scalar],
    cell_size: int):
  """Calculate n-step Q-learning loss for pixel control auxiliary task.

  For each pixel-based pseudo reward signal, the corresponding action-value
  function is trained off-policy, using Q(lambda). A discount of 0.9 is
  commonly used for learning the value functions.

  Note that, since pseudo rewards have a spatial structure, with neighbouring
  cells exhibiting strong correlations, it is convenient to predict the action
  values for all the cells through a deconvolutional head.

  See "Reinforcement Learning with Unsupervised Auxiliary Tasks" by Jaderberg,
  Mnih, Czarnecki et al. (https://arxiv.org/abs/1611.05397).

  Args:
    observations: A tensor of shape `[T+1, ...]`; `...` is the observation
      shape, `T` the sequence length.
    actions: A tensor, shape `[T,]`, of the actions across each sequence.
    action_values: A tensor, shape `[T+1, H, W, N]` of pixel control action
      values, where `H`, `W` are the number of pixel control cells/tasks, and
      `N` is the number of actions.
    discount_factor: discount used for learning the value function associated
      to the pseudo rewards; must be a scalar or a Tensor of shape [T].
    cell_size: size of the cells used to derive the pixel based pseudo-rewards.

  Returns:
    a tensor containing the spatial loss, shape [T, H, W].

  Raises:
    ValueError: if the shape of `action_values` is not compatible with that of
      the pseudo-rewards derived from the observations.
  """
  # Check shapes
  assert len(actions.shape) == 1
  assert len(action_values.shape) == 4
  # Check types
  chex.assert_type([observations], float)
  chex.assert_type([actions], int)
  # Useful shapes.
  sequence_length = actions.shape[0]
  num_actions = action_values.shape[-1]
  height_width_q = action_values.shape[1:-1]
  # Calculate rewards using the observations.
  # Compute pseudo-rewards and get their shape.
  pseudo_rewards = general_value_functions.pixel_control_rewards(
      observations, cell_size)
  height_width = pseudo_rewards.shape[1:]
  # Check that pseudo-rewards and Q-values are compatible in shape.
  if height_width != height_width_q:
    raise ValueError(
        "Pixel Control values are not compatible with the shape of the"
        "pseudo-rewards derived from the observation. Pseudo-rewards have "
        f"shape {height_width}, while Pixel Control values have "
        f"shape {height_width_q}")
  # We now have Q(s,a) and rewards, so can calculate the n-step loss. The
  # QLambda loss op expects inputs of shape [T,N] and [T], but our tensors
  # are in a variety of incompatible shapes. The state-action values have
  # shape [T,H,W,N] and rewards have shape [T,H,W]. We can think of the
  # [H,W] dimensions as extra batch dimensions for the purposes of the loss
  # calculation, so we first collapse [H,W] into a single dimension.
  q_tm1 = jnp.reshape(action_values[:-1], (sequence_length, -1, num_actions))
  r_t = jnp.reshape(pseudo_rewards, (sequence_length, -1))
  q_t = jnp.reshape(action_values[1:], (sequence_length, -1, num_actions))
  # The actions tensor is of shape [T], and is the same for each H and W.
  # We thus expand it to be same shape as the reward tensor, [T,HW].
  expanded_actions = actions[..., None, None]
  a_tm1 = jnp.tile(expanded_actions, (1,) + height_width)
  a_tm1 = jnp.reshape(a_tm1, (sequence_length, -1))
  # We similarly expand-and-tile the discount to [T,HW].
  discount_factor = jnp.asarray(discount_factor)
  if not discount_factor.shape:
    pcont_t = jnp.reshape(discount_factor, (1,))
    pcont_t = jnp.tile(pcont_t, a_tm1.shape)
  elif len(discount_factor.shape) == 1:
    tiled_pcont = jnp.tile(
        discount_factor[:, None, None], (1,) + height_width)
    pcont_t = jnp.reshape(tiled_pcont, (sequence_length, -1))
  else:
    raise ValueError(
        "The discount_factor must be a scalar or a tensor of rank 1. "
        f"instead is a tensor of shape {discount_factor.shape}")
  # Compute a QLambda loss of shape [T,HW]
  batched_q_lambda = jax.vmap(
      functools.partial(
          value_learning.q_lambda, lambda_=1.0),
      in_axes=1, out_axes=1)
  td_error = batched_q_lambda(q_tm1, a_tm1, r_t, pcont_t, q_t)
  loss = 0.5 * td_error**2
  expanded_shape = (sequence_length,) + height_width
  spatial_loss = jnp.reshape(loss, expanded_shape)  # [T,H,W].
  return spatial_loss


def expectile_loss(predictions: Array, targets: Array, expectile: float):
  """Expectile loss that weighs over-/under-estimations of targets differently.

  The formulation is an asymmetric least squares loss which downweights the
  contributions of predictions that underestimate targets and upweights
  overestimation of targets.
  See Implicit Q Learning (https://arxiv.org/pdf/2110.06169.pdf) Sec 4.1 for
  more details on this loss.

  Args:
    predictions: Tensor of shape [T, ...] where T is the sequence length.
    targets: Target values for predictions with identical shape as predictions.
    expectile: A float value that represents the weights assigned to predictions
      that over-estimate targets.

  Returns:
    a vector of same shape as predictions.
  """
  chex.assert_equal_shape([predictions, targets])
  diff = (targets - predictions)
  is_underestimation = jnp.less(diff, 0)
  weight = jnp.abs(expectile - is_underestimation)
  return weight * (diff**2)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities to construct and learn from policy targets."""

import functools

import chex
import distrax
import jax
import jax.numpy as jnp


@chex.dataclass(frozen=True)
class PolicyTarget:
  """A dataclass to hold (possibly sampled) policy targets."""
  # The sampled target actions. These may not cover all actions.
  # The shape is [N_targets, ...]
  actions: chex.Array
  # Probabilities for the corresponding actions. They may have been
  # importance weighted. The shape matches log_prob(actions).shape.
  weights: chex.Array


def constant_policy_targets(
    distribution: distrax.DistributionLike,
    rng_key: chex.PRNGKey,
    num_samples: int,
    weights_scale: float = 1.) -> PolicyTarget:
  """Create policy targets with constant weights.

  The actions will be sampled from `distribution` and will all be associated to
  a constant `weights_scale`. If `distribution` is a (discrete or continuous)
  uniform probability distribution, distilling these targets will push the
  agent's policy towards a uniform distribution. The strength of the penalty
  associated with a non-uniform policy depends on `weights_scale` (e.g. in the
  extreme case `weights_scale==0`, distillation loss is 0 for any policy).

  Args:
    distribution: a `distrax` or `tfp` distribution for sampling actions.
    rng_key: a JAX pseudo random number generator.
    num_samples: number of action sampled in the PolicyTarget.
    weights_scale: constant weight associated with each action.

  Returns:
    a PolicyTarget to learn from.
  """
  random_actions = distribution.sample(
      seed=rng_key, sample_shape=(num_samples,))
  return PolicyTarget(
      actions=random_actions,
      weights=weights_scale * jnp.ones(
          (num_samples,) + distribution.batch_shape))


zero_policy_targets = functools.partial(
    constant_policy_targets, weights_scale=0.0)


def sampled_policy_distillation_loss(
    distribution: distrax.DistributionLike,
    policy_targets: PolicyTarget,
    stop_target_gradients: bool = True,
) -> chex.Numeric:
  """Compute a sampled cross-entropy-like loss.

  The loss corresponds to taking the mean of `-weights * log_prob(actions)`,
  where the weights and actions come from a `PolicyTarget` object, and the mean
  is computed over the N samples in the `policy_target` as well as any batch
  dimension. The loss is suitable for both discrete and continuous actions.

  Args:
    distribution: a predicted `distrax` or `tfp` distribution.
    policy_targets: a policy target to learn from.
    stop_target_gradients: bool indicating whether or not to apply a stop
      gradient to the policy_targets, default True.

  Returns:
    a scalar loss.
  """
  # Optionally, stop gradients from propagating into the targets and into
  # the actions; the latter is mostly relevant in continuous control.
  weights = jax.lax.select(
      stop_target_gradients,
      jax.lax.stop_gradient(policy_targets.weights), policy_targets.weights)
  actions = jax.lax.select(
      stop_target_gradients,
      jax.lax.stop_gradient(policy_targets.actions), policy_targets.actions)
  # Compute log-probabilities.
  log_probs = distribution.log_prob(actions)
  # Assert shapes are compatible.
  chex.assert_equal_shape([weights, log_probs])
  # We avoid NaNs from `0 * (-inf)` by using `0 * min_logp` in that case.
  min_logp = jnp.finfo(log_probs.dtype).min
  # We average over the samples, over time and batch, and if the actions are
  # a continuous vector also over the actions.
  return -jnp.mean(weights * jnp.maximum(log_probs, min_logp))
